# -*- coding: utf-8 -*-

"""
Forked from https://github.com/coudertlab/elate/
Modified:
    - 
"""

import json
import math
import os
import platform
import random
import re
import sys
import time

from collections import OrderedDict
from io import StringIO
import requests

import numpy as np
from scipy import optimize


__author__ = "Romain Gaillac and François-Xavier Coudert"
__version__ = "2022.06.14"
__license__ = "MIT"


def removeHTMLTags(s):
    """Remove HTML tags, notably for use as page title"""
    return re.sub('<[^<]+?>', '', s)


def finishWebPage(outbuffer):
    """ Write the footer and finish the page """

    print('<div id="footer" class="content">')
    print('Code version: ' + __version__ + ' (running on Python ' + platform.python_version() + ')<br/>')
    print('<script type="text/javascript">var endTime = %g;' % time.perf_counter())
    print('document.write("Execution time: " + (endTime-startTime).toFixed(3) + " seconds<br/>");')
    print('if(typeof isOrtho !== \'undefined\') document.write("Specific (faster) code for orthorhombic case was used.");')
    print('</script></div>')
    print('</div>')
    print('</body></html>')
    return outbuffer.getvalue()


def writeHeader(outbuffer, title="Elastic Tensor Analysis"):
    """ Write the header of the HTML page """

    print("""
    <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">
    <html>
    <head>
        <title>%s</title>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
        <link rel="stylesheet" type="text/css" href="/default.css" />
        <link rel="stylesheet" type="text/css" href="https://cdn.jsdelivr.net/npm/jsxgraph@1.1.0/distrib/jsxgraph.css" />
        <script src="https://cdn.jsdelivr.net/npm/jsxgraph@1.1.0/distrib/jsxgraphcore.js"></script>
        <script src="http://cdn.plot.ly/plotly-latest.min.js"></script>
        <script src="https://cdnjs.cloudflare.com/ajax/libs/jquery/1.12.4/jquery.min.js"></script>
        </head>
    """ % (title))


# printTitle writes the introduction of Elate
def printTitle(outbuffer, title="Elastic Tensor Analysis"):
    writeHeader(outbuffer, title)
    print("""
        <body>

        <div class="content">
        <h1><a href="/elate">ELATE: Elastic tensor analysis</a></h1>

        <p>Welcome to ELATE, the online tool for analysis of elastic tensors, developed by <b>Romain Gaillac</b> and <b><a
        href="http://coudert.name">François-Xavier Coudert</a></b> at <a href="http://www.chimie-paristech.fr/molsim/">CNRS / Chimie
        ParisTech</a>. <br/> If you use the software in published results (paper, conference, etc.), please cite the <a
        href="http://dx.doi.org/10.1088/0953-8984/28/27/275201">corresponding paper</a> (<em>J. Phys. Condens. Matter</em>, 2016, 28, 275201) and give the
        website URL.</p>

        <p>ELATE is <a href="https://github.com/fxcoudert/elate">open source software</a>. Any queries or comments are welcome at
    <script type="text/javascript">
    //<![CDATA[
    var c_="";for(var o5=0;o5<411;o5++)c_+=String.fromCharCode(("s%oz65j5>oJ.~~vs!Kt00}.~|}{\\"$s~%}!s0Kv#\\"wv<s!~tjjK{j5wo#zH}<j5s!z~qo6s~=u=i:00ikk>97a6!#|w<u!t{}vQ!o}Qsr?6F8G9:B8D9>@?7>a9!#|w<u!t{}vQ!o}QsrB67Dj59}qr$!s8#vq{wsw~;!oAA\\"wA#qsj5v!<~sozsq=6=A:u00970i0<ikk>a9!#|w<u!t{}vQ!o}QsrA69DDD>:E\\'7@<7s!z~qo6sjj==8:uN070j59j5jj.0|}}{\\"$}s#$0Kv#\\"wv<s!Ktj5jjj5jjL0\\'t14>O>>DBqI$}sr#!14>>>>BDqIwvw{sO~;!o\\"ws#vq14>>B>ID!t=JLo<j5s!z~qo6sO=u=0:705<!s~zoqs6=6<76<7=u:02@2?07<\\"$p\\"#!6?77".charCodeAt(o5)-(14)+0x3f)%(2*6+83)+64-32);document.write(eval(c_))
    //]]>
    </script>
        </p>
    """)


# 3D plot functions
################################################################################################

def write3DPlotData(dataX, dataY, dataZ, dataR, n, opacity=1.0):

    showcont = "true"
    if (opacity != 1.0):
        showcont = "false"
    if (n == 1):
        js = OrderedDict([
            ("x", dataX),
            ("y", dataY),
            ("z", dataZ),
            ("text", dataR),
            ("showscale", "false"),
            ("colorscale", "[[\'0\',\'rgb(22,136,51)\'],[\'0.125\',\'rgb(61,153,85)\'],[\'0.25\',\'rgb(121,178,136)\'],[\'0.375\',\'rgb(181,204,187)\'],[\'0.5\',\'rgb(195,230,200)\'],[\'0.625\',\'rgb(181,204,187)\'],[\'0.75\',\'rgb(121,178,136)\'],[\'0.875\',\'rgb(61,153,85)\'],[\'1\',\'rgb(22,136,51)\']]"),
            ("zsmooth", "'fast'"),
            ("type", "'surface'"),
            ("hoverinfo", "'text'"),
            ("opacity", opacity),
            ("contours", "{x :{ show:"+showcont+", color: 'rgb(192,192,192)'},y :{ show:"+showcont+", color: 'rgb(192,192,192)'},z :{ show:"+showcont+", color: 'rgb(192,192,192)'}}")
        ])

    if (n == 2):
        js = OrderedDict([
            ("x", dataX),
            ("y", dataY),
            ("z", dataZ),
            ("text", dataR),
            ("showscale", "false"),
            ("colorscale", "[[\'0\',\'rgb(180,4,38)\'],[\'0.125\',\'rgb(222,96,77)\'],[\'0.25\',\'rgb(244,154,123)\'],[\'0.375\',\'rgb(245,196,173)\'],[\'0.5\',\'rgb(246,216,201)\'],[\'0.625\',\'rgb(245,196,173)\'],[\'0.75\',\'rgb(244,154,123)\'],[\'0.875\',\'rgb(222,96,77)\'],[\'1\',\'rgb(180,4,38)\']]"),
            ("zsmooth", "'fast'"),
            ("type", "'surface'"),
            ("hoverinfo", "'text'"),
            ("opacity", opacity),
            ("contours", "{x :{ show:"+showcont+", color: 'rgb(192,192,192)'},y :{ show:"+showcont+", color: 'rgb(192,192,192)'},z :{ show:"+showcont+", color: 'rgb(192,192,192)'}}")
        ])

    if (n == 3):
        js = OrderedDict([
            ("x", dataX),
            ("y", dataY),
            ("z", dataZ),
            ("text", dataR),
            ("showscale", "false"),
            ("colorscale", "[[\'0\',\'rgb(59,76,192)\'],[\'0.125\',\'rgb(98,130,234)\'],[\'0.25\',\'rgb(141,176,254)\'],[\'0.375\',\'rgb(184,208,249)\'],[\'0.5\',\'rgb(207,223,250)\'],[\'0.625\',\'rgb(184,208,249)\'],[\'0.75\',\'rgb(141,176,254)\'],[\'0.875\',\'rgb(98,130,234)\'],[\'1\',\'rgb(59,76,192)\']]"),
            ("zsmooth", "'fast'"),
            ("type", "'surface'"),
            ("hoverinfo", "'text'"),
            ("opacity", opacity),
            ("contours", "{x :{ show:"+showcont+", color: 'rgb(192,192,192)'},y :{ show:"+showcont+", color: 'rgb(192,192,192)'},z :{ show:"+showcont+", color: 'rgb(192,192,192)'}}")
        ])

    print(json.dumps(js, indent=3).replace('\"', '') + ";")


def make3DPlot(func, legend='', width=600, height=600, npoints=200):

    str1 = legend.split("\'")[0]
    str2 = legend.split("\'")[1]

    u = np.linspace(0, np.pi, npoints)
    v = np.linspace(0, 2*np.pi, 2*npoints)
    r = np.zeros(len(u)*len(v))

    dataX = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataY = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataZ = [[0.0 for i in range(len(v))] for j in range(len(u))]
    dataR = [["0.0" for i in range(len(v))] for j in range(len(u))]

    count = 0
    for cu in range(len(u)):
        for cv in range(len(v)):
            r_tmp = func(u[cu], v[cv])
            z = r_tmp * np.cos(u[cu])
            x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
            y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
            dataX[cu][cv] = x
            dataY[cu][cv] = y
            dataZ[cu][cv] = z
            dataR[cu][cv] = "'E = "+str(float(int(10*r_tmp))/10.0)+" GPa, "+"\u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
            count = count+1

    i = random.randint(0, 100000)
    print('<div class="plot3D">')
    print('<div id="box%d" style="width: %dpx; height: %dpx; display:block;"></div>' % (i, width, height))
    print('</div>')
    print('<script type="text/javascript">')
    print("var trace =")
    write3DPlotData(dataX, dataY, dataZ, dataR, 1)
    print("var data = [trace]")
    print("var layout =")
    layout = {"title": "\'"+str1+"\\"+"\'"+str2+"\'", "width": "650", "height": "700", "autosize": "false", "autorange": "true", "margin": "{l: 65, r: 50, b: 65, t: 90}"}
    print(json.dumps(layout, indent=3).replace('\\\\', '\\').replace('\"', '') + ";")
    print("Plotly.newPlot('box%d',data,layout);" % (i))
    print('</script>')


def make3DPlotPosNeg(func, legend='', width=600, height=600, npoints=200):

  u = np.linspace(0, np.pi, npoints)
  v = np.linspace(0, 2*np.pi, 2*npoints)

  dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  for cu in range(len(u)):
    for cv in range(len(v)):
      r_tmp = max(0, func(u[cu], v[cv]))
      z = r_tmp * np.cos(u[cu])
      x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
      y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
      dataX1[cu][cv] = x
      dataY1[cu][cv] = y
      dataZ1[cu][cv] = z
      dataR1[cu][cv] = "'"+"\u03B2 = "+str(float(int(10*r_tmp))/10.0)+" TPa'"+"+'-1'.sup()+"+"', \u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
      count = count+1

  dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  for cu in range(len(u)):
    for cv in range(len(v)):
      r_tmp = max(0, -func(u[cu], v[cv]))
      z = r_tmp * np.cos(u[cu])
      x = r_tmp * np.sin(u[cu]) * np.cos(v[cv])
      y = r_tmp * np.sin(u[cu]) * np.sin(v[cv])
      dataX2[cu][cv] = x
      dataY2[cu][cv] = y
      dataZ2[cu][cv] = z
      dataR2[cu][cv] = "'"+"\u03B2 = -"+str(float(int(10*r_tmp))/10.0)+" TPa'"+"+'-1'.sup()+"+"', \u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
      count = count+1

  i = random.randint(0, 100000)
  print('<div class="plot3D">')
  print('<div id="box%d" style="width: %dpx; height: %dpx; display:block;"></div>' % (i, width, height))
  print('</div>')
  print('<script type="text/javascript">')
  print("var trace1 =")
  write3DPlotData(dataX1, dataY1, dataZ1, dataR1, 1)
  print("var trace2 =")
  write3DPlotData(dataX2, dataY2, dataZ2, dataR2, 2)
  print("var data = [trace1, trace2]")
  print("var layout =")
  layout = {"title": "\'"+legend+"\'", "width": "650", "height": "700", "autosize": "false", "autorange": "true", "margin": "{l: 65, r: 50, b: 65, t: 90}"}
  print(json.dumps(layout, indent=3).replace('\\\\', '\\').replace('\"', '') + ";")
  print("Plotly.newPlot('box%d',data,layout);" % (i))
  print('</script>')


def make3DPlot2(func, legend='', width=600, height=600, npoints=50):

  u = np.linspace(0, np.pi, npoints)
  v = np.linspace(0, np.pi, npoints)
  w = [v[i]+np.pi for i in range(1,len(v))]
  v = np.append(v, w)

  dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  r = [0.0,0.0,np.pi/2.0,np.pi/2.0]
  for cu in range(len(u)):
    for cv in range(len(v)):

      r = func(u[cu],v[cv],r[2],r[3])
      z = np.cos(u[cu])
      x = np.sin(u[cu]) * np.cos(v[cv])
      y = np.sin(u[cu]) * np.sin(v[cv])

      r1_tmp = r[0]
      z1 = r1_tmp * z
      x1 = r1_tmp * x
      y1 = r1_tmp * y
      dataX1[cu][cv] = x1
      dataY1[cu][cv] = y1
      dataZ1[cu][cv] = z1
      dataR1[cu][cv] = "'"+"G'"+"+'min'.sub()+"+"' = "+str(float(int(10*r1_tmp))/10.0)+"GPa, "+"\u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"

      r2_tmp = r[1]
      z2 = r2_tmp * z
      x2 = r2_tmp * x
      y2 = r2_tmp * y
      dataX2[cu][cv] = x2
      dataY2[cu][cv] = y2
      dataZ2[cu][cv] = z2
      dataR2[cu][cv] = "'"+"G'"+"+'max'.sub()+"+"' = "+str(float(int(10*r1_tmp))/10.0)+"GPa, "+"\u03B8 = "+str(float(int(10*u[cu]*180/np.pi))/10.0)+"\u00B0, "+"\u03c6 = "+str(float(int(10*v[cv]*180/np.pi))/10.0)+"\u00B0'"
      count = count+1

  i = random.randint(0, 100000)
  print('<div class="plot3D">')
  print('<div id="box%d" style="width: %dpx; height: %dpx; display:block;"></div>' % (i, width, height))
  print('</div>')
  print('<script type="text/javascript">')
  print("var trace1 =")
  write3DPlotData(dataX1, dataY1, dataZ1, dataR1, 1)
  print("var trace2 =")
  write3DPlotData(dataX2, dataY2, dataZ2, dataR2, 3, 0.5)
  print("var data = [trace1, trace2]")
  print("var layout =")
  layout = {"title": "\'"+legend+"\'", "width":"650", "height":"700" , "autosize":"false", "autorange":"true", "margin": "{l: 65, r: 50, b: 65, t: 90}"}
  print(json.dumps(layout, indent=3).replace('\\\\','\\').replace('\"','') + ";")
  print("Plotly.newPlot('box%d',data,layout);" % (i))
  print('</script>')


def make3DPlot3(func, legend='', width=600, height=600, npoints=50):

  str1 = legend.split("\'")[0]
  str2 = legend.split("\'")[1]

  u = np.linspace(0, np.pi, npoints)
  v = np.linspace(0, np.pi, npoints)
  w = [v[i]+np.pi for i in range(1,len(v))]
  v = np.append(v, w)

  dataX1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ1 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR1 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  dataX2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ2 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR2 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  dataX3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataY3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataZ3 = [[0.0 for i in range(len(v))] for j in range(len(u))]
  dataR3 = [["0.0" for i in range(len(v))] for j in range(len(u))]

  count = 0
  r = [0.0, 0.0, 0.0, np.pi/2.0, np.pi/2.0]
  ruv = [[r for i in range(len(u))] for j in range(len(v))]
  for cu in range(len(u)):
    for cv in range(len(v)):
       ruv[cv][cu] = func(u[cu],v[cv],r[3],r[4])

  for cu in range(len(u)):
    for cv in range(len(v)):

      z = np.cos(u[cu])
      x = np.sin(u[cu]) * np.cos(v[cv])
      y = np.sin(u[cu]) * np.sin(v[cv])

      r = ruv[cv][cu]
      r1_tmp = r[0]
      dataX1[cu][cv] = r1_tmp * x
      dataY1[cu][cv] = r1_tmp * y
      dataZ1[cu][cv] = r1_tmp * z
      dataR1[cu][cv] = "'"+"\u03BD'"+"+'min'.sub()+"+"' = "+str(float(int(100*r1_tmp))/100.0)+", "+"\u03B8 = "+str(float(int(100*u[cu]*180/np.pi))/100.0)+"\u00B0, "+"\u03c6 = "+str(float(int(100*v[cv]*180/np.pi))/100.0)+"\u00B0'"

      r2_tmp = r[1]
      dataX2[cu][cv] = r2_tmp * x
      dataY2[cu][cv] = r2_tmp * y
      dataZ2[cu][cv] = r2_tmp * z
      dataR2[cu][cv] = "'"+"\u03BD'"+"+'min'.sub()+"+"' = "+str(float(int(100*r2_tmp))/100.0)+", "+"\u03B8 = "+str(float(int(100*u[cu]*180/np.pi))/100.0)+"\u00B0, "+"\u03c6 = "+str(float(int(100*v[cv]*180/np.pi))/100.0)+"\u00B0'"

      r3_tmp = r[2]
      dataX3[cu][cv] = r3_tmp * x
      dataY3[cu][cv] = r3_tmp * y
      dataZ3[cu][cv] = r3_tmp * z
      dataR3[cu][cv] = "'"+"\u03BD'"+"+'max'.sub()+"+"' = "+str(float(int(100*r3_tmp))/100.0)+", "+"\u03B8 = "+str(float(int(100*u[cu]*180/np.pi))/100.0)+"\u00B0, "+"\u03c6 = "+str(float(int(100*v[cv]*180/np.pi))/100.0)+"\u00B0'"
      count = count+1

  i = random.randint(0, 100000)
  print('<div class="plot3D">')
  print('<div id="box%d" style="width: %dpx; height: %dpx; display:block;"></div>' % (i, width, height))
  print('</div>')
  print('<script type="text/javascript">')
  print("var trace1 =")
  write3DPlotData(dataX1, dataY1, dataZ1, dataR1, 2, 0.5)
  print("var trace2 =")
  write3DPlotData(dataX2, dataY2, dataZ2, dataR2, 1, 1.0)
  print("var trace3 =")
  write3DPlotData(dataX3, dataY3, dataZ3, dataR3, 3, 0.5)
  print("var data = [trace1, trace2, trace3]")
  print("var layout =")
  layout = {"title": "\'"+str1+"\\"+"\'"+str2+"\'", "width":"650", "height":"700" , "autosize":"false", "autorange":"true", "margin": "{l: 65, r: 50, b: 65, t: 90}"}
  print(json.dumps(layout, indent=3).replace('\\\\','\\').replace('\"','') + ";")
  print("Plotly.newPlot('box%d',data,layout);" % (i))
  print('</script>')




# Polar plot functions
################################################################################################

def writePolarPlotData(dataX, dataY, suffix):
    """Write data for a polar plot, taking care of the center of inversion"""

    print("var dataX" + suffix + " = [")
    print((len(dataX) * "%.5f,") % tuple(dataX))
    print(((len(dataX)-1) * "%.5f," + "%.5f") % tuple(-dataX))
    print("];")
    print("var dataY" + suffix + " = [")
    print((len(dataX) * "%.5f,") % tuple(dataY))
    print(((len(dataX)-1) * "%.5f," + "%.5f") % tuple(-dataY))
    print("];")



def makePolarPlot(func, maxrad, legend='', p='xy', width=300, height=300, npoints=90, color='#009010', linewidth=2):

    i = random.randint(0, 100000)
    print('<div class="plot">')
    print('<div id="box%d" class="jxgbox" style="width: %dpx; height: %dpx; display:inline-block;"></div>' % (i, width, height))
    print('<br />%s</div>' % legend)
    print('<script type="text/javascript">')
    print('var b = JXG.JSXGraph.initBoard(\'box%d\', {boundingbox: [-%f, %f, %f, -%f], axis:true, showcopyright: 0});'
          % (i, maxrad, maxrad, maxrad, maxrad))

    u = np.linspace(0, np.pi, npoints)
    r = list(map(func, u))
    if (p=="xy"):
        x = r * np.cos(u)
        y = r * np.sin(u)
    else:
        y = r * np.cos(u)
        x = r * np.sin(u)

    writePolarPlotData (x, y, "")
    print("b.create('curve', [dataX,dataY], {strokeColor:'%s', strokeWidth: %d});" % (color, linewidth))
    print('</script>')

def makePolarPlotPosNeg(func, maxrad, legend='', p='xy', width=300, height=300, npoints=90, linewidth=2):
    i = random.randint(0, 100000)
    print('<div class="plot">')
    print('<div id="box%d" class="jxgbox" style="width: %dpx; height: %dpx; display:inline-block;"></div>' % (i, width, height))
    print('<br />%s</div>' % legend)
    print('<script type="text/javascript">')
    print('var b = JXG.JSXGraph.initBoard(\'box%d\', {boundingbox: [-%f, %f, %f, -%f], axis:true, showcopyright: 0});'
          % (i, maxrad, maxrad, maxrad, maxrad))

    u = np.linspace(0, np.pi, npoints)
    r = list(map(lambda x: max(0, func(x)), u))
    if (p=="xy"):
        x1 = r * np.cos(u)
        y1 = r * np.sin(u)
    else:
        y1 = r * np.cos(u)
        x1 = r * np.sin(u)
    r = list(map(lambda x: max(0, -func(x)), u))
    if (p=="xy"):
        x2 = r * np.cos(u)
        y2 = r * np.sin(u)
    else:
        y2 = r * np.cos(u)
        x2 = r * np.sin(u)

    writePolarPlotData (x1, y1, "1")
    writePolarPlotData (x2, y2, "2")
    print("b.create('curve', [dataX1,dataY1], {strokeColor:'green', strokeWidth: %d});" % (linewidth))
    print("b.create('curve', [dataX2,dataY2], {strokeColor:'red', strokeWidth: %d});" % (linewidth))
    print('</script>')

def makePolarPlot2(func, maxrad, legend='', p='xy', width=300, height=300, npoints=61, linewidth=2):
    i = random.randint(0, 100000)
    print('<div class="plot">')
    print('<div id="box%d" class="jxgbox" style="width: %dpx; height: %dpx; display:inline-block;"></div>' % (i, width, height))
    print('<br />%s</div>' % legend)
    print('<script type="text/javascript">')
    print('var b = JXG.JSXGraph.initBoard(\'box%d\', {boundingbox: [-%f, %f, %f, -%f], axis:true, showcopyright: 0});'
          % (i, maxrad, maxrad, maxrad, maxrad))

    u = np.linspace(0, np.pi, npoints)
    r = list(map(func, u))

    if (p=="xy"):
        x1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
        y1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
        x2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
        y2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])
    else:
        y1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
        x1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
        y2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
        x2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])

    writePolarPlotData (x1, y1, "1")
    writePolarPlotData (x2, y2, "2")
    print("b.create('curve', [dataX1,dataY1], {strokeColor:'green', strokeWidth: %d});" % (linewidth))
    print("b.create('curve', [dataX2,dataY2], {strokeColor:'blue', strokeWidth: %d});" % (linewidth))
    print('</script>')

def makePolarPlot3(func, maxrad, legend='', p='xy', width=300, height=300, npoints=61, linewidth=2):
    i = random.randint(0, 100000)
    print('<div class="plot">')
    print('<div id="box%d" class="jxgbox" style="width: %dpx; height: %dpx; display:inline-block;"></div>' % (i, width, height))
    print('<br />%s</div>' % legend)
    print('<script type="text/javascript">')
    print('var b = JXG.JSXGraph.initBoard(\'box%d\', {boundingbox: [-%f, %f, %f, -%f], axis:true, showcopyright: 0});'
          % (i, maxrad, maxrad, maxrad, maxrad))

    u = np.linspace(0, np.pi, npoints)
    r = list(map(func, u))

    if (p=="xy"):
        x1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
        y1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
        x2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
        y2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])
        x3 = np.array([ ir[2] * np.cos(iu) for ir, iu in zip(r,u) ])
        y3 = np.array([ ir[2] * np.sin(iu) for ir, iu in zip(r,u) ])
    else:
        y1 = np.array([ ir[0] * np.cos(iu) for ir, iu in zip(r,u) ])
        x1 = np.array([ ir[0] * np.sin(iu) for ir, iu in zip(r,u) ])
        y2 = np.array([ ir[1] * np.cos(iu) for ir, iu in zip(r,u) ])
        x2 = np.array([ ir[1] * np.sin(iu) for ir, iu in zip(r,u) ])
        y3 = np.array([ ir[2] * np.cos(iu) for ir, iu in zip(r,u) ])
        x3 = np.array([ ir[2] * np.sin(iu) for ir, iu in zip(r,u) ])

    writePolarPlotData (x1, y1, "1")
    writePolarPlotData (x2, y2, "2")
    writePolarPlotData (x3, y3, "3")
    print("b.create('curve', [dataX1,dataY1], {strokeColor:'red', strokeWidth: %d});" % (linewidth))
    print("b.create('curve', [dataX2,dataY2], {strokeColor:'green', strokeWidth: %d});" % (linewidth))
    print("b.create('curve', [dataX3,dataY3], {strokeColor:'blue', strokeWidth: %d});" % (linewidth))
    print('</script>')


################################################################################################

def dirVec(theta, phi):
    return [ math.sin(theta)*math.cos(phi), math.sin(theta)*math.sin(phi), math.cos(theta) ]

def dirVec1(theta, phi, chi):
    return [ math.sin(theta)*math.cos(phi), math.sin(theta)*math.sin(phi), math.cos(theta) ]

def dirVec2(theta, phi, chi):
    return [ math.cos(theta)*math.cos(phi)*math.cos(chi) - math.sin(phi)*math.sin(chi),
             math.cos(theta)*math.sin(phi)*math.cos(chi) + math.cos(phi)*math.sin(chi),
             - math.sin(theta)*math.cos(chi) ]


# Functions to minimize/maximize
def minimize(func, dim):
    if dim == 2:
        r = ((0, np.pi), (0, np.pi))
        n = 25
    elif dim == 3:
        r = ((0, np.pi), (0, np.pi), (0, np.pi))
        n = 10

    # TODO -- try basin hopping or annealing
    return optimize.brute(func, r, Ns = n, full_output = True, finish = optimize.fmin)[0:2]


def maximize(func, dim):
    res = minimize(lambda x: -func(x), dim)
    return (res[0], -res[1])


class Elastic:
    """An elastic tensor, along with methods to access it"""

    def __init__(self, s):
        """Initialize the elastic tensor from a string"""

        if not s:
            raise ValueError("no matrix was provided")

        # Argument can be a 6-line string, a list of list, or a string representation of the list of list
        try:
            if type(json.loads(s)) == list: s = json.loads(s)
        except:
            pass

        if type(s) == str:
            # Remove braces and pipes
            s = s.replace("|", " ").replace("(", " ").replace(")", " ")

            # Remove empty lines
            lines = [line for line in s.split('\n') if line.strip()]
            if len(lines) != 6:
                raise ValueError("should have six rows")

            # Convert to float
            try:
                mat = [list(map(float, line.split())) for line in lines]
            except:
                raise ValueError("not all entries are numbers")
        elif type(s) == list:
            # If we already have a list, simply use it
            mat = s
        else:
            raise ValueError("invalid argument as matrix")

        # Make it into a square matrix
        mat = np.array(mat)
        if mat.shape != (6,6):
            # Is it upper triangular?
            if list(map(len, mat)) == [6,5,4,3,2,1]:
                mat = [ [0]*i + mat[i] for i in range(6) ]
                mat = np.array(mat)

        # Is it lower triangular?
        if list(map(len, mat)) == [1,2,3,4,5,6]:
            mat = [ mat[i] + [0]*(5-i) for i in range(6) ]
            mat = np.array(mat)

        if mat.shape != (6,6):
            raise ValueError("should be a square matrix")

        # Check that is is symmetric, or make it symmetric
        if np.linalg.norm(np.tril(mat, -1)) == 0:
            mat = mat + np.triu(mat, 1).transpose()
        if np.linalg.norm(np.triu(mat, 1)) == 0:
            mat = mat + np.tril(mat, -1).transpose()
        if np.linalg.norm(mat - mat.transpose()) > 1e-3:
            raise ValueError("should be symmetric, or triangular")
        elif np.linalg.norm(mat - mat.transpose()) > 0:
            mat = 0.5 * (mat + mat.transpose())

        # Store it
        self.CVoigt = mat

        # Put it in a more useful representation
        try:
            self.SVoigt = np.linalg.inv(self.CVoigt)
        except:
            raise ValueError("matrix is singular")

        VoigtMat = [[0, 5, 4], [5, 1, 3], [4, 3, 2]]
        def SVoigtCoeff(p,q): return 1. / ((1+p//3)*(1+q//3))

        self.Smat = [[[[ SVoigtCoeff(VoigtMat[i][j], VoigtMat[k][l]) * self.SVoigt[VoigtMat[i][j]][VoigtMat[k][l]]
                         for i in range(3) ] for j in range(3) ] for k in range(3) ] for l in range(3) ]
        return

    def isOrthorhombic(self):
        def iszero(x): return (abs(x) < 1.e-3)
        return (iszero(self.CVoigt[0][3]) and iszero(self.CVoigt[0][4]) and iszero(self.CVoigt[0][5])
                and iszero(self.CVoigt[1][3]) and iszero(self.CVoigt[1][4]) and iszero(self.CVoigt[1][5])
                and iszero(self.CVoigt[2][3]) and iszero(self.CVoigt[2][4]) and iszero(self.CVoigt[2][5])
                and iszero(self.CVoigt[3][4]) and iszero(self.CVoigt[3][5]) and iszero(self.CVoigt[4][5]))

    def isCubic(self):
        def iszero(x): return (abs(x) < 1.e-3)
        return (iszero(self.CVoigt[0][3]) and iszero(self.CVoigt[0][4]) and iszero(self.CVoigt[0][5])
                and iszero(self.CVoigt[1][3]) and iszero(self.CVoigt[1][4]) and iszero(self.CVoigt[1][5])
                and iszero(self.CVoigt[2][3]) and iszero(self.CVoigt[2][4]) and iszero(self.CVoigt[2][5])
                and iszero(self.CVoigt[3][4]) and iszero(self.CVoigt[3][5]) and iszero(self.CVoigt[4][5])
                and iszero(self.CVoigt[0][0] - self.CVoigt[1][1]) and iszero(self.CVoigt[0][0] - self.CVoigt[2][2])
                and iszero(self.CVoigt[0][0] - self.CVoigt[1][1]) and iszero(self.CVoigt[0][0] - self.CVoigt[2][2])
                and iszero(self.CVoigt[3][3] - self.CVoigt[4][4]) and iszero(self.CVoigt[3][3] - self.CVoigt[5][5])
                and iszero(self.CVoigt[0][1] - self.CVoigt[0][2]) and iszero(self.CVoigt[0][1] - self.CVoigt[1][2]))

    def Young(self, x):
        a = dirVec(x[0], x[1])
        r = sum([ a[i]*a[j]*a[k]*a[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return 1/r

    def Young_2(self,x,y):
        a = dirVec(x, y)
        r = sum([ a[i]*a[j]*a[k]*a[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return 1/r

    def LC(self, x):
        a = dirVec(x[0], x[1])
        r = sum([ a[i]*a[j] * self.Smat[i][j][k][k]
                  for i in range(3) for j in range(3) for k in range(3) ])
        return 1000 * r

    def LC_2(self, x, y):
        a = dirVec(x, y)
        r = sum([ a[i]*a[j] * self.Smat[i][j][k][k]
                  for i in range(3) for j in range(3) for k in range(3) ])
        return 1000 * r

    def shear(self, x):
        a = dirVec(x[0], x[1])
        b = dirVec2(x[0], x[1], x[2])
        r = sum([ a[i]*b[j]*a[k]*b[l] * self.Smat[i][j][k][l]
                  for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return 1/(4*r)

    def Poisson(self, x):
        a = dirVec(x[0], x[1])
        b = dirVec2(x[0], x[1], x[2])
        r1 = sum([ a[i]*a[j]*b[k]*b[l] * self.Smat[i][j][k][l]
                   for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        r2 = sum([ a[i]*a[j]*a[k]*a[l] * self.Smat[i][j][k][l]
                   for i in range(3) for j in range(3) for k in range(3) for l in range(3) ])
        return -r1/r2

    def averages(self):
        A = (self.CVoigt[0][0] + self.CVoigt[1][1] + self.CVoigt[2][2]) / 3
        B = (self.CVoigt[1][2] + self.CVoigt[0][2] + self.CVoigt[0][1]) / 3
        C = (self.CVoigt[3][3] + self.CVoigt[4][4] + self.CVoigt[5][5]) / 3
        a = (self.SVoigt[0][0] + self.SVoigt[1][1] + self.SVoigt[2][2]) / 3
        b = (self.SVoigt[1][2] + self.SVoigt[0][2] + self.SVoigt[0][1]) / 3
        c = (self.SVoigt[3][3] + self.SVoigt[4][4] + self.SVoigt[5][5]) / 3

        KV = (A + 2*B) / 3
        GV = (A - B + 3*C) / 5

        KR = 1 / (3*a + 6*b)
        GR = 5 / (4*a - 4*b + 3*c)

        KH = (KV + KR) / 2
        GH = (GV + GR) / 2

        return [ [KV, 1/(1/(3*GV) + 1/(9*KV)), GV, (1 - 3*GV/(3*KV+GV))/2],
                 [KR, 1/(1/(3*GR) + 1/(9*KR)), GR, (1 - 3*GR/(3*KR+GR))/2],
                 [KH, 1/(1/(3*GH) + 1/(9*KH)), GH, (1 - 3*GH/(3*KH+GH))/2] ]

    def shear2D(self, x):
        ftol = 0.001
        xtol = 0.01
        def func1(z): return self.shear([x[0], x[1], z])
        r1 = optimize.minimize(func1, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.shear([x[0], x[1], z])
        r2 = optimize.minimize(func2, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun))

    def shear3D(self, x, y, guess1 = np.pi/2.0, guess2 = np.pi/2.0):
        tol = 0.005
        def func1(z): return self.shear([x, y, z])
        r1 = optimize.minimize(func1, guess1, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.shear([x, y, z])
        r2 = optimize.minimize(func2, guess2, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        return (float(r1.fun), -float(r2.fun), float(r1.x), float(r2.x))

    def Poisson2D(self, x):
        ftol = 0.001
        xtol = 0.01
        def func1(z): return self.Poisson([x[0], x[1], z])
        r1 = optimize.minimize(func1, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.Poisson([x[0], x[1], z])
        r2 = optimize.minimize(func2, np.pi/2.0, args=(), method = 'Powell', options={"xtol":xtol, "ftol":ftol})#, bounds=[(0.0,np.pi)])
        return (min(0,float(r1.fun)), max(0,float(r1.fun)), -float(r2.fun))

    def poisson3D(self, x, y, guess1 = np.pi/2.0, guess2 = np.pi/2.0):
        tol = 0.005
        def func1(z): return self.Poisson([x, y, z])
        r1 = optimize.minimize(func1, guess1, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        def func2(z): return -self.Poisson([x, y, z])
        r2 = optimize.minimize(func2, guess2, args=(), method = 'COBYLA', options={"tol":tol})#, bounds=[(0.0,np.pi)])
        return (min(0,float(r1.fun)), max(0,float(r1.fun)), -float(r2.fun), float(r1.x), float(r2.x))


class ElasticOrtho(Elastic):
    """An elastic tensor, for the specific case of an orthorhombic system"""

    def __init__(self, arg):
        """Initialize from a matrix, or from an Elastic object"""
        if type(arg) == str:
            Elastic.__init__(self, arg)
        elif isinstance(arg, Elastic):
            self.CVoigt = arg.CVoigt
            self.SVoigt = arg.SVoigt
            self.Smat = arg.Smat
        else:
            raise TypeError("ElasticOrtho constructor argument should be string or Elastic object")

    def Young(self, x):
        ct2 = math.cos(x[0])**2
        st2 = 1 - ct2
        cf2 = math.cos(x[1])**2
        sf2 = 1 - cf2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        return 1/(ct2**2*s33 + 2*cf2*ct2*s13*st2 + cf2*ct2*s55*st2 + 2*ct2*s23*sf2*st2 + ct2*s44*sf2*st2 + cf2**2*s11*st2**2 + 2*cf2*s12*sf2*st2**2 + cf2*s66*sf2*st2**2 + s22*sf2**2*st2**2)

    def LC(self, x):
        ct2 = math.cos(x[0])**2
        cf2 = math.cos(x[1])**2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        return 1000 * (ct2 * (s13 + s23 + s33) + (cf2 * (s11 + s12 + s13) + (s12 + s22 + s23) * (1 - cf2)) * (1 - ct2))

    def shear(self, x):
        ct = math.cos(x[0])
        ct2 = ct*ct
        st2 = 1 - ct2
        cf = math.cos(x[1])
        sf = math.sin(x[1])
        sf2 = sf*sf
        cx = math.cos(x[2])
        cx2 = cx*cx
        sx = math.sin(x[2])
        sx2 = 1 - cx2
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]
        r = (
            ct2*ct2*cx2*s44*sf2 + cx2*s44*sf2*st2*st2 + 4*cf**3*ct*cx*(-2*s11 + 2*s12 + s66)*sf*st2*sx
            + 2*cf*ct*cx*sf*(ct2*(s44 - s55) + (4*s13 - 4*s23 - s44 + s55 - 4*s12*sf2 + 4*s22*sf2 - 2*s66*sf2)*st2)*sx
            + s66*sf2*sf2*st2*sx2 + cf**4*st2*(4*ct2*cx2*s11 + s66*sx2)
            + ct2*(2*cx2*(2*s33 + sf2*(-4*s23 - s44 + 2*s22*sf2))*st2 + s55*sf2*sx2)
            + cf**2*(ct2*ct2*cx2*s55 + ct2*(-2*cx2*(4*s13 + s55 - 2*(2*s12 + s66)*sf2)*st2 + s44*sx2)
                     + st2*(cx2*s55*st2 + 2*(2*s11 - 4*s12 + 2*s22 - s66)*sf2*sx2))
            )
        return 1/r

    def Poisson(self, x):
        ct = math.cos(x[0])
        ct2 = ct*ct
        st2 = 1 - ct2
        cf = math.cos(x[1])
        sf = math.sin(x[1])
        cx = math.cos(x[2])
        sx = math.sin(x[2])
        s11 = self.Smat[0][0][0][0]
        s22 = self.Smat[1][1][1][1]
        s33 = self.Smat[2][2][2][2]
        s44 = 4 * self.Smat[1][2][1][2]
        s55 = 4 * self.Smat[0][2][0][2]
        s66 = 4 * self.Smat[0][1][0][1]
        s12 = self.Smat[0][0][1][1]
        s13 = self.Smat[0][0][2][2]
        s23 = self.Smat[1][1][2][2]

        return (
    (-(ct**2*cx**2*s33*st2) - cf**2*cx**2*s13*st2*st2 - cx**2*s23*sf**2*st2*st2 + ct*cx*s44*sf*st2*(ct*cx*sf + cf*sx) -
            ct**2*s23*(ct*cx*sf + cf*sx)**2 - cf**2*s12*st2*(ct*cx*sf + cf*sx)**2 - s22*sf**2*st2*(ct*cx*sf + cf*sx)**2 +
            cf*ct*cx*s55*st2*(cf*ct*cx - sf*sx) - cf*s66*sf*st2*(ct*cx*sf + cf*sx)*(cf*ct*cx - sf*sx) -
            ct**2*s13*(cf*ct*cx - sf*sx)**2 - cf**2*s11*st2*(cf*ct*cx - sf*sx)**2 - s12*sf**2*st2*(cf*ct*cx - sf*sx)**2)/
            (ct**4*s33 + 2*cf**2*ct**2*s13*st2 + cf**2*ct**2*s55*st2 + 2*ct**2*s23*sf**2*st2 + ct**2*s44*sf**2*st2 +
            cf**4*s11*st2*st2 + 2*cf**2*s12*sf**2*st2*st2 + cf**2*s66*sf**2*st2*st2 + s22*sf**4*st2*st2)
        )


################################################################################################

# Materials Project URL
urlBase = 'https://legacy.materialsproject.org/rest'


def queryMaterials(query, mapiKey):
    """Return a list of material IDs for a given query string"""

    # If the query is a material ID, return it
    if query[0:3] == "mp-": return [query]

    try:
        r = requests.get(f'{urlBase}/v2/materials/{query}/mids', headers={"X-API-KEY": mapiKey})
        resp = r.json()
    except Exception as e:
        print(str(e), file=sys.stderr)
        return []

    if (not resp['valid_response']): return []
    return resp['response']


def queryElasticityV2(mat, mapiKey):
    """Return elastic properties for a given material ID, using V2 MAPI"""

    data = { 'criteria': '{"task_id": "' + mat + '"}',
             'properties': '["formula", "pretty_formula", "material_id", "elasticity"]',
             'API_KEY': mapiKey }
    try:
        r = requests.post(f'{urlBase}/v2/query', data)
        resp = r.json()
    except Exception as e:
        print(str(e), file=sys.stderr)
        return None

    if not resp["valid_response"]: return None
    if len(resp["response"]) > 1: raise(Exception("Multiple results returned"))
    if len(resp["response"]) == 0: return None
    return resp["response"][0]


def ELATE_MaterialsProject(query, mapiKey):
  """Call ELATE with a query from the Materials Project"""

  # If we were directly given a material ID, or there is a simple match
  materials = queryMaterials(query, mapiKey)
  if len(materials) == 1:
    r = queryElasticityV2(query, mapiKey)

    if r and 'elasticity' in r:
      tensor = r["elasticity"]["elastic_tensor"]
      return ELATE(tensor, '%s (Materials Project id <a href="%s%s" target="_blank">%s</a>)' % (r["pretty_formula"], "https://www.materialsproject.org/materials/", r["material_id"], r["material_id"]))

  # Otherwise, run the MP query, list the matches and let the user choose
  sys.stdout = outbuffer = StringIO()
  printTitle(outbuffer, "ELATE: Elastic tensor analysis")
  print('<h2>Query from the Materials Project database</h2>')

  # Either there was no match, or a single match with no elastic data
  if len(materials) <= 1:
    print("""<p>
            Your query for <tt style="background-color: #e0e0e0;">%s</tt> from the <a href="https://materialsproject.org">Materials Project</a> database
            has returned a total of zero result. Or is it zero results? In any case, we are very sorry.</p>

            <p>If you wish, you can try another query here:
            <form name="elastic" action="/elate/mp" method="get">
              <input type="text" name="query" style="font-family:sans-serif; width: 20em;">
              <input type="submit" style="font-size: 100%%; color: #b02020;" value="Submit query">
            </form>
            or go back to our <a href="/elate">main page</a>.
            </p>""" % (query))
    return finishWebPage(outbuffer)


  print("""<p>Your query for <tt style="background-color: #e0e0e0;">%s</tt> from the <a href="https://materialsproject.org">Materials Project</a> database
           has returned %d results.""" % (query, len(materials)))

  if len(materials) > 10:
    materials = materials[0:10]
    print("Below is a table of the 10 first matches.")

  print("<table><tr><th>Identifier</th><th>Formula</th><th>Elastic data</th></tr>")
  for mat in materials:
    r = queryElasticityV2(mat, mapiKey)
    print('<tr><td><a href="https://www.materialsproject.org/materials/%s" target="_blank">%s</a></td><td>%s</td>' % (mat, mat, r["pretty_formula"]))
    if "elasticity" in r and r["elasticity"]:
      print('<td>Elastic data available, <a href="/elate/mp?%s" target="_blank">perform analysis</a></td></tr>' % (mat))
    else:
      print('<td>No elastic data available</td></tr>')
  print("</table>")

  return finishWebPage(outbuffer)


def ELATE(matrix, sysname):
  """ELATE performs the calculation and plots every property in 2D"""

  # Redirect output to out string buffer
  sys.stdout = outbuffer = StringIO()

  # Start timing
  print('<script type="text/javascript">var startTime = %g</script>' % time.perf_counter())

  printTitle(outbuffer, "Elastic analysis of " + removeHTMLTags(sysname))

  try:
    elas = Elastic(matrix)
  except ValueError as e:
    print('<div class="error">Invalid stiffness matrix: ')
    print(e.args[0])
    if matrix:
      print('<pre>' + str(matrix) + '</pre>')

    print('</div>')
    print('<input action="action" type="button" value="Go back" onclick="window.history.go(-1); return false;" />')
    return finishWebPage(outbuffer)

  if elas.isOrthorhombic():
    elas = ElasticOrtho(elas)
    print('<script type="text/javascript">var isOrtho = 1;</script>')

  print('<h2>Summary of the properties</h2>')

  print('<h3>Input: stiffness matrix (coefficients in GPa) of %s</h3>' % (sysname))
  print('<pre>')
  for i in range(6):
    print(("   " + 6*"%7.5g  ") % tuple(elas.CVoigt[i]))
  print('</pre>')

  avg = elas.averages()
  print('<h3>Average properties</h3>')

  print("<table><tr><th>Averaging scheme</th><th>Bulk modulus</th><th>Young's modulus</th><th>Shear modulus</th><th>Poisson's ratio</th></tr>")
  print(('<tr><td>Voigt</td><td><em>K</em><sub>V</sub> = %7.5g GPa</td><td><em>E</em><sub>V</sub> = %7.5g GPa</td>'
          + '<td><em>G</em><sub>V</sub> = %7.5g GPa</td><td><em>&nu;</em><sub>V</sub> = %.5g</td></tr>')
        % tuple(avg[0]))
  print(('<tr><td>Reuss</td><td><em>K</em><sub>R</sub> = %7.5g GPa</td><td><em>E</em><sub>R</sub> = %7.5g GPa</td>'
          + '<td><em>G</em><sub>R</sub> = %7.5g GPa</td><td><em>&nu;</em><sub>R</sub> = %.5g</td></tr>')
        % tuple(avg[1]))
  print(('<tr><td>Hill</td><td><em>K</em><sub>H</sub> = %7.5g GPa</td><td><em>E</em><sub>H</sub> = %7.5g GPa</td>'
          + '<td><em>G</em><sub>H</sub> = %7.5g GPa</td><td><em>&nu;</em><sub>H</sub> = %.5g</td></tr>')
        % tuple(avg[2]))
  print('</table>')


  print('''<h3>Eigenvalues of the stiffness matrix</h3>
  <table><tr>
  <th>&lambda;<sub>1</sub></th>
  <th>&lambda;<sub>2</sub></th>
  <th>&lambda;<sub>3</sub></th>
  <th>&lambda;<sub>4</sub></th>
  <th>&lambda;<sub>5</sub></th>
  <th>&lambda;<sub>6</sub></th>
  </tr><tr>''')
  eigenval = sorted(np.linalg.eig(elas.CVoigt)[0])
  print((6*'<td>%7.5g GPa</td>') % tuple(eigenval))
  print('</tr></table>')

  if eigenval[0] <= 0:
    print('<div class="error">Stiffness matrix is not definite positive, crystal is mechanically unstable<br/>')
    print('No further analysis will be performed.</div>')
    return finishWebPage(outbuffer)


  minE = minimize(elas.Young, 2)
  maxE = maximize(elas.Young, 2)
  minLC = minimize(elas.LC, 2)
  maxLC = maximize(elas.LC, 2)
  minG = minimize(elas.shear, 3)
  maxG = maximize(elas.shear, 3)
  minNu = minimize(elas.Poisson, 3)
  maxNu = maximize(elas.Poisson, 3)

  print("""<h3>Variations of the elastic moduli</h3>
            <table>
            <tr><td></td><th colspan="2">Young\'s modulus</th><th colspan="2">Linear compressibility</th>
            <th colspan="2">Shear modulus</th><th colspan="2">Poisson\'s ratio</th><th></th></tr>
            <td></td><th><em>E</em><sub>min</sub></th><th><em>E</em><sub>max</sub></th>
            <th>&beta;<sub>min</sub></th><th>&beta;<sub>max</sub></th><th><em>G</em><sub>min</sub></th><th><em>G</em><sub>max</sub></th>
            <th>&nu;<sub>min</sub></th><th>&nu;<sub>max</sub></th><th></th></tr>""")

  print(('<tr><td>Value</td><td>%8.5g GPa</td><td>%8.5g GPa</td>'
        + '<td>%8.5g TPa<sup>&ndash;1</sup></td><td>%8.5g TPa<sup>&ndash;1</sup></td>'
        + '<td>%8.5g GPa</td><td>%8.5g GPa</td>'
        + '<td>%.5g</td><td>%.5g</td><td>Value</td></tr>') % (minE[1], maxE[1], minLC[1], maxLC[1], minG[1], maxG[1], minNu[1], maxNu[1]))

  anisE = '%8.4g' % (maxE[1]/minE[1])
  anisLC = ('%8.4f' % (maxLC[1]/minLC[1])) if minLC[1] > 0 else "&infin;"
  anisG = '%8.4g' % (maxG[1]/minG[1])
  anisNu = ('%8.4f' % (maxNu[1]/minNu[1])) if minNu[1]*maxNu[1] > 0 else "&infin;"
  print(('<tr><td>Anisotropy</td>' + 4 * '<td colspan="2">%s</td>'
        + '<td>Anisotropy</td></tr>') % (anisE, anisLC, anisG, anisNu))

  print('<tr><td>Axis</td>')
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec(*minE[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec(*maxE[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec(*minLC[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec(*maxLC[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec1(*minG[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec1(*maxG[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec1(*minNu[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec1(*maxNu[0])))
  print('<td>Axis</td></tr>')

  print('<tr><td></td><td></td><td></td><td></td><td></td>')
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec2(*minG[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec2(*maxG[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec2(*minNu[0])))
  print('<td>%.4f<br />%.4f<br />%.4f</td>' % tuple(dirVec2(*maxNu[0])))
  print('<td>Second axis</td></tr></table>')

  print("<h2>Spatial dependence of Young's modulus</h2>")
  print("""<form id="elastic" action="/wait3D" method="post" target="_blank">
            <textarea name="matrix" style="display: none;">%s</textarea>
            <textarea name="sysname" style="display: none;">%s</textarea>
            <textarea name="job" style="display: none;">%s</textarea>
            <br /><input type="submit" style="font-size: 100%%; color: #b02020;" value="Visualize in 3D">
           </form>""" % (matrix, sysname, "young"))
  m = 1.2 * maxE[1]
  makePolarPlot(lambda x: elas.Young([np.pi / 2, x]), m, "Young's modulus in (xy) plane", "xy")
  makePolarPlot(lambda x: elas.Young([x, 0]), m, "Young's modulus in (xz) plane", "xz")
  makePolarPlot(lambda x: elas.Young([x, np.pi / 2]), m, "Young's modulus in (yz) plane", "yz")

  print("<h2>Spatial dependence of linear compressibility</h2>")
  print("""<form id="elastic" action="/wait3D" method="post" target="_blank">
            <textarea name="matrix" style="display: none;">%s</textarea>
            <textarea name="sysname" style="display: none;">%s</textarea>
            <textarea name="job" style="display: none;">%s</textarea>
            <br /><input type="submit" style="font-size: 100%%; color: #b02020;" value="Visualize in 3D">
           </form>""" % (matrix, sysname, "lc"))
  m = 1.2 * max(maxLC[1], abs(minLC[1]))
  makePolarPlotPosNeg(lambda x: elas.LC([np.pi / 2, x]), m, "linear compressibility in (xy) plane", "xy")
  makePolarPlotPosNeg(lambda x: elas.LC([x, 0]), m, "linear compressibility in (xz) plane", "xz")
  makePolarPlotPosNeg(lambda x: elas.LC([x, np.pi / 2]), m, "linear compressibility in (yz) plane", "yz")

  print("<h2>Spatial dependence of shear modulus</h2>")
  print("""<form id="elastic" action="/wait3D" method="post" target="_blank">
            <textarea name="matrix" style="display: none;">%s</textarea>
            <textarea name="sysname" style="display: none;">%s</textarea>
            <textarea name="job" style="display: none;">%s</textarea>
            <br /><input type="submit" style="font-size: 100%%; color: #b02020;" value="Visualize in 3D">
           </form>""" % (matrix, sysname, "shear"))
  m = 1.2 * maxG[1]
  makePolarPlot2(lambda x: elas.shear2D([np.pi / 2, x]), m, "Shear modulus in (xy) plane", "xy")
  makePolarPlot2(lambda x: elas.shear2D([x, 0]), m, "Shear modulus in (xz) plane", "xz")
  makePolarPlot2(lambda x: elas.shear2D([x, np.pi / 2]), m, "Shear modulus in (yz) plane", "yz")

  print("<h2>Spatial dependence of Poisson's ratio</h2>")
  print("""<form id="elastic" action="/wait3D" method="post" target="_blank">
            <textarea name="matrix" style="display: none;">%s</textarea>
            <textarea name="sysname" style="display: none;">%s</textarea>
            <textarea name="job" style="display: none;">%s</textarea>
            <br /><input type="submit" style="font-size: 100%%; color: #b02020;" value="Visualize in 3D">
           </form>""" % (matrix, sysname, "poisson"))
  m = 1.2 * max(abs(maxNu[1]), abs(minNu[1]))
  makePolarPlot3(lambda x: elas.Poisson2D([np.pi / 2, x]), m, "Poisson's ratio in (xy) plane", "xy")
  makePolarPlot3(lambda x: elas.Poisson2D([x, 0]), m, "Poisson's ratio in (xz) plane", "xz")
  makePolarPlot3(lambda x: elas.Poisson2D([x, np.pi / 2]), m, "Poisson's ratio in (yz) plane", "yz")

  print("</div>")
  return finishWebPage(outbuffer)


def wait3D(matrix, sysname, job):
    """Display a waiting page while we calculate a 3D plot"""

    sys.stdout = outbuffer = StringIO()
    writeHeader(outbuffer, "Young 3D for " + removeHTMLTags(sysname))

    print("""
    <div class="content">
    <img src="/loading.gif" alt="[loading]" />
    <p>Please wait while your 3D graph is loading… (it can take from 15 seconds up to a minute)</p>
    """)

    # Pass arguments
    print("""
    <form id="elastic" action="/plot3D" method="post" style="display: none;">
        <textarea name="matrix">%s</textarea>
        <textarea name="sysname">%s</textarea>
        <textarea name="job">%s</textarea>
        <input type="submit" value="">
    </form>""" % (matrix, sysname, job))

    # Reload immediately
    print("""
    <script type="text/javascript">
        window.onload = function(){
        setTimeout(function () {
            document.getElementById("elastic").submit();
                }, 100);
            };
    </script>""")

    return finishWebPage(outbuffer)


def plot3D(matrix, sysname, job):
    """Display a 3D plot"""

    # Dispatch to the specific function depending on type
    functions = {'young': YOUNG3D, 'lc': LC3D, 'shear': SHEAR3D, 'poisson': POISSON3D}
    return functions[job](matrix, sysname)


# ELATE : basic usage of the tool, only 2D plots
# YOUNG3D : visualize Young's modulus in 3D
# LC3D : visualize Linear compressiblity in 3D
# SHEAR3D : visualize Shear modulus in 3D
# POISSON3D : visualize Poisson ratio in 3D
################################################################################################


def YOUNG3D(matrix, sysname):

    sys.stdout = outbuffer = StringIO()
    writeHeader(outbuffer, "Young 3D for " + removeHTMLTags(sysname))

    # Start timing
    print('<script type="text/javascript">var startTime = %g</script>' % time.perf_counter())
    print('<div class="content">')

    print("<h1> 3D Visualization of Young's modulus </h1>")
    elas = Elastic(matrix)
    if elas.isOrthorhombic():
        elas = ElasticOrtho(elas)
        print('<script type="text/javascript">var isOrtho = 1;</script>')

    make3DPlot(lambda x, y: elas.Young_2(x, y), "Young's modulus")

    print('<h3>Input: stiffness matrix (coefficients in GPa) of %s</h3>' % (sysname))
    print('<pre>')
    for i in range(6):
        print(("   " + 6 * "%7.5g  ") % tuple(elas.CVoigt[i]))
    print('</pre></div>')
    return finishWebPage(outbuffer)


def LC3D(matrix, sysname):

    sys.stdout = outbuffer = StringIO()
    writeHeader(outbuffer, "LC 3D for " + removeHTMLTags(sysname))

    # Start timing
    print('<script type="text/javascript">var startTime = %g</script>' % time.perf_counter())
    print('<div class="content">')

    print("<h1> 3D Visualization of Linear compressiblity </h1>")
    elas = Elastic(matrix)
    if elas.isOrthorhombic():
        elas = ElasticOrtho(elas)
        print('<script type="text/javascript">var isOrtho = 1;</script>')

    make3DPlotPosNeg(lambda x, y: elas.LC_2(x, y), "Linear compressiblity")

    print('<h3>Input: stiffness matrix (coefficients in GPa) of %s</h3>' % (sysname))
    print('<pre>')
    for i in range(6):
        print(("   " + 6 * "%7.5g  ") % tuple(elas.CVoigt[i]))
    print('</pre></div>')
    return finishWebPage(outbuffer)


def SHEAR3D(matrix, sysname):

    sys.stdout = outbuffer = StringIO()
    writeHeader(outbuffer, "Shear 3D for " + removeHTMLTags(sysname))

    # Start timing
    print('<script type="text/javascript">var startTime = %g</script>' % time.perf_counter())
    print('<div class="content">')

    print("<h1> 3D Visualization of Shear modulus </h1>")
    elas = Elastic(matrix)
    if elas.isOrthorhombic():
        elas = ElasticOrtho(elas)
        print('<script type="text/javascript">var isOrtho = 1;</script>')

    make3DPlot2(lambda x, y, g1, g2: elas.shear3D(x, y, g1, g2), "Shear modulus")

    print('<h3>Input: stiffness matrix (coefficients in GPa) of %s</h3>' % (sysname))
    print('<pre>')
    for i in range(6):
        print(("   " + 6 * "%7.5g  ") % tuple(elas.CVoigt[i]))
    print('</pre></div>')
    return finishWebPage(outbuffer)


def POISSON3D(matrix, sysname):

    sys.stdout = outbuffer = StringIO()
    writeHeader(outbuffer, "Poisson 3D for " + removeHTMLTags(sysname))

    # Start timing
    print('<script type="text/javascript">var startTime = %g</script>' % time.perf_counter())
    print('<div class="content">')

    print("<h1> 3D Visualization of Poisson's ratio </h1>")
    elas = Elastic(matrix)
    if elas.isOrthorhombic():
        elas = ElasticOrtho(elas)
        print('<script type="text/javascript">var isOrtho = 1;</script>')

    make3DPlot3(lambda x, y, g1, g2: elas.poisson3D(x, y, g1, g2), "Poisson's ratio")

    print('<h3>Input: stiffness matrix (coefficients in GPa) of %s</h3>' % (sysname))
    print('<pre>')
    for i in range(6):
        print(("   " + 6 * "%7.5g  ") % tuple(elas.CVoigt[i]))
    print('</pre></div>')
    return finishWebPage(outbuffer)



"""
Module containing everything related to plotting: 
- hvplot extra functions
"""

import hvplot
from bokeh.io import export_svg
import holoviews as hv  
from cairosvg import svg2png

import amof.files.path 

def save_hvplot(plot, filename, format = 'svg+png'):
    """
    Save hvplot to filename

    Args:
        plot: hvplot object
        filename: string or pathlib object indicating where to save output
        format: str, can be 'svg+png' (high res png) or 'svg' or 'png' (low res)
    """
    if format[0:3] == 'svg':
        output_filename = str(amof.files.path.append_suffix(filename, 'svg'))
        bp = hv.render(plot)
        bp.output_backend = "svg"
        export_svg(bp, filename=output_filename)
        if format == 'svg+png':
            svg_code = open(output_filename, 'rt').read()
            png_filename = str(amof.files.path.append_suffix(filename, 'png'))
            svg2png(bytestring=svg_code,write_to=png_filename, scale = 3) # default is 100
    elif format == 'png':
        output_filename = str(amof.files.path.append_suffix(filename, 'png'))
        hvplot.save(plot, output_filename)
    else:
        raise ValueError('Format not supported')


"""
Module containing path related methods
"""

import pathlib

def append_suffix(path, suffix):
    """
    Append suffix to path if suffix is not already path latest suffix

    Args:
        path: pathlib.Path object or string
        suffix: string object, if no initial '.' it is added
    Returns:
        path: pathlib.Path object
    """
    if len(suffix)!=0 and suffix[0]!='.':
        suffix = '.' + suffix
    path = pathlib.Path(path)
    if path.suffix != suffix:
        path = path.parent / (path.name + suffix)
    return path


"""
Module containing msd related methods
"""

import os

# force numpy to use one thread 
os.environ["NUMEXPR_NUM_THREADS"] = "1"  # seems to suffice

import ase
import ase.data
from ase.geometry.geometry import wrap_positions

import numpy as np
import pandas as pd
import logging
import joblib

import amof.trajectory as amtraj
import amof.atom
import amof.files.path

logger = logging.getLogger(__name__)

class Msd(object):
    """
    Main class for MSD
    """

    def write_to_file(self, path_to_output):
        """path_to_output: where the MSD object will be written"""
        path_to_output = amof.files.path.append_suffix(path_to_output, 'msd')
        self.data.to_feather(path_to_output)

    @classmethod
    def from_msd(cls, *args):
        logger.exception('from_msd is deprecated, use from_file instead')

    @classmethod
    def from_file(cls, path_to_msd):
        """
        constructor of msd class from msd file
        """
        msd_class = cls() # initialize class
        msd_class.read_msd_file(path_to_msd)
        return msd_class # return class as it is a constructor

    def read_msd_file(self, path_to_data):
        """path_to_data: where the MSD object is"""
        path_to_data = amof.files.path.append_suffix(path_to_data, 'msd')
        self.data = pd.read_feather(path_to_data)


class DirectMsd(Msd):
    """
    Direct MSD

    MSD(t) = \dfrac{1}{N_{particles}} \sum_{i=1}^{N_{particles}} (r_i(t) - r_i(0))^2 

    This precise implementation only works for orthogonal cell.
    Better to use WindowMSD
    """

    def __init__(self):
        """default constructor"""
        self.data = pd.DataFrame({"Step": np.empty([0])})
        logger.warning('DirectMsd is deprecated and not suitable for non-orthogonal cells, use WindowMsd instead')

    @classmethod
    def from_trajectory(cls, trajectory, delta_Step = 1, first_frame = 0, parallel = False):
        """
        constructor of msd class

        Args:
            trajectory: ase trajectory object
            delta_Step: number of simulation steps between two frames
            parallel: Boolean or int (number of cores to use): whether to parallelize the computation
        """
        msd_class = cls() # initialize class
        step = amtraj.construct_step(delta_Step=delta_Step, first_frame = first_frame, number_of_frames = len(trajectory))
        msd_class.compute_msd(trajectory, step, parallel)
        return msd_class # return class as it is a constructor

    @staticmethod
    def compute_species_msd(trajectory, atomic_number = None):
        """calculate MSD with real pos (stored in r) compared to PBC pos stored
        in ase (extracted with position)
        if atomic_number is None, compute MSD between all atoms
        """
        r_0 = amof.atom.select_species_positions(trajectory[0], atomic_number)
        r_t = r_0
        MSD = np.zeros(len(trajectory))
        for t in range(1, len(trajectory)):
            r_t_minus_dt = r_t
            dr = np.zeros((len(r_0), 3))
            # ! only works for orthogonal cell
            for j in range(3): #x,y,z
                a = trajectory[t].get_cell()[j,j]
                dr[:,j] = (amof.atom.select_species_positions(trajectory[t], atomic_number) - r_t_minus_dt%a)[:,j]
                for i in range(len(dr)):
                    if dr[i][j]>a/2:
                        dr[i][j] -= a
                    elif dr[i][j]<-a/2:
                        dr[i][j] += a
            r_t = dr + r_t_minus_dt
            MSD[t] = np.linalg.norm(r_t-r_0)**2/len(r_0)
        return MSD

    def compute_msd(self, trajectory, step, parallel):
        """
        Args:
            trajectory: ase trajectory object
            step: np array, simulation steps
            parallel: Boolean or int (number of cores to use): whether to parallelize the computation
        """
        logger.info("Start computing msd for %s frames", len(trajectory))
        
        elements = amof.atom.get_atomic_numbers_unique(trajectory[0])

        Step = step     
        self.data = pd.DataFrame({"Step": Step})
        if not parallel:
            self.data["X"] = self.compute_species_msd(trajectory)

            for x in elements:
                x_str = ase.data.chemical_symbols[x]
                self.data[x_str] = self.compute_species_msd(trajectory, x)
        else:
            x_list = [None] + elements
            num_cores = min(len(x_list), joblib.cpu_count()) # default value
            if type(parallel) == int and parallel < num_cores:
                num_cores = parallel
            msd_list = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(self.compute_species_msd)(trajectory, x) for x in x_list)
            self.data["X"] = msd_list[0]
            for i in range(1, len(x_list)):
                x_str = ase.data.chemical_symbols[x_list[i]]
                self.data[x_str] = msd_list[i]               


class WindowMsd(Msd):
    """
    Window MSD

    MSD(m) = \frac{1}{N_{particles}} \sum_{i=1}^{N_{particles}} \frac{1}{N-m} \sum_{k=0}^{N-m-1} (\vec{r}_i(k+m) - \vec{r}_i(k))^2
    m is the window length size, usually replaced by time t in publications. 
    To be unambiguous, here time will carry units and m and k will be used as trajectory indexes
    The final data array will be y = MSD, x = m labelled 'Time' 

    Time is expressed in fs
    """

    def __init__(self):
        """default constructor"""
        self.data = pd.DataFrame({"Time": np.empty([0])})

    @classmethod
    def from_trajectory(cls, trajectory, delta_time = 100, max_time = "half", timestep = 1, parallel = False):
        """
        constructor of msd class

        Args:
            trajectory: ase trajectory object
            delta_time: int, time between two computed values of the MSD, in fs
            max_time: int or str
                if half, then the upper limit is half of the simulation size
                if input max_time is higher than half of simulation size, will choose the latter
            timestep: int, time between two frames of the trajectory
            parallel: Boolean or int (number of cores to use): whether to parallelize the computation
        """
        msd_class = cls() # initialize class
        half_time = (len(trajectory) // 2) * timestep
        if max_time == "half" or max_time > half_time:
            max_time = half_time
        if delta_time < timestep:
            logger.exception("Delta_time should be larger than timestep")
        delta_m = delta_time // timestep
        window = np.arange(0, max_time // timestep, delta_m)
        time = timestep * window
        msd_class.compute_msd(trajectory, window, time, parallel)
        return msd_class # return class as it is a constructor

    @staticmethod
    def compute_msd_of_m(delta_pos, m):
        """calculate MSD(m) for a given m with real pos (stored in r) compared to PBC pos stored
        in ase (extracted with position)

        Args:
            delta_pos: atomic coordinates of successive displacements
            m: int
            if atomic_number is None, compute MSD between all atoms
        """
        MSD_partial = np.zeros(len(delta_pos) - m)
        r_k_minus_m = delta_pos[0]
        r_k = r_k_minus_m * 0 # empty array with same shape as r_k_minus_m
        for k in range(0, m+1): 
            r_k += delta_pos[k]
        for k in range(m+1, len(delta_pos)): 
            r_k += delta_pos[k]
            r_k_minus_m += delta_pos[k-m]
            MSD_partial[k - m] = np.linalg.norm(r_k - r_k_minus_m)**2/len(r_k_minus_m)
        MSD = np.mean(MSD_partial)
        return MSD

    def compute_msd(self, trajectory, window, time, parallel):
        """
        Args:
            trajectory: ase trajectory object
            window: np array, window
            time: np array, time in fs
            parallel: Boolean or int (number of cores to use): whether to parallelize the computation
        """
        logger.info("Start computing msd at %s times on a trajectory of %s frames", len(window), len(trajectory))
        
        elements = amof.atom.get_atomic_numbers_unique(trajectory[0])

        # Compensate MD drift of center of mass
        for atom in trajectory:
            cg = atom.get_center_of_mass()
            atom.translate(-cg)

        positions_by_elt = [] # each partial trajectory consist of the po
        for x in elements:
            x_str = ase.data.chemical_symbols[x]
            positions_by_elt.append([amof.atom.select_species_positions(atom, x) for atom in trajectory])
        cell = [atom.get_cell() for atom in trajectory]

        self.data = pd.DataFrame({"Time": time}) 

        def compute_for_every_m(positions, cell):
            delta_pos = amtraj.get_delta_pos(positions, cell)
            return [self.compute_msd_of_m(delta_pos, m) for m in window]

        if not parallel:
            msd_list = [compute_for_every_m(pos, cell) for pos in positions_by_elt]
        else:
            num_cores = len(elements) # default value
            if type(parallel) == int and parallel < num_cores:
                num_cores = parallel
            msd_list = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(compute_for_every_m)(pos, cell) for pos in positions_by_elt)

        # assign partial msd
        for i in range(len(elements)):
            x_str = ase.data.chemical_symbols[elements[i]]
            self.data[x_str] = msd_list[i]
        # compute total msd
        formula_dict = trajectory[0].symbols.formula._count
        self.data['X'] = self.data.groupby(self.data.index).apply(
            lambda x: 
                np.sum([x[k] * v for k, v in formula_dict.items()])
                / sum(formula_dict.values())
                )         



"""
Module containing bond angle distribution related methods
"""
import os

# force numpy to use one thread 
os.environ["OMP_NUM_THREADS"] = "1"  # essential
os.environ["OPENBLAS_NUM_THREADS"] = "1"  # at least one of these 4 is needed
os.environ["MKL_NUM_THREADS"] = "1"  
os.environ["NUMEXPR_NUM_THREADS"] = "1"  
os.environ["OPENBLAS_MAIN_FREE"] = "1"

import ase
import ase.neighborlist
import ase.data

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import xarray as xr
import joblib
import itertools

import logging

import amof.trajectory
import amof.files.path
import amof.atom as amatom

# create logger without parameters for this module file that will be incorporated by the main file logging parameters
logger = logging.getLogger(__name__)

class CoreBad(object):
    """
    Core material for every Bad function
    """

    @classmethod
    def from_trajectory(cls, trajectory, nb_set_and_cutoff, dtheta = 0.05, normalization = 'total', parallel = False, ):
        """
        constructor of bad class from an ase trajectory object
        Args:
            nb_set_and_cutoff: dict, keys are str indicating pair of neighbours, 
                values are cutoffs float, in Angstrom
            dtheta: float, in degrees, default value of 0.05degree similar to RG
            normalization: str, can be 'total' (entire bad is normalized) or 'partial' (sum of partial is normalized)
        """
        bad_class = cls() # initialize class
        bad_class.compute_bad(trajectory, nb_set_and_cutoff, dtheta, normalization, parallel)
        return bad_class # return class as it is a constructor

    @classmethod
    def from_file(cls, filename):
        """
        constructor of bad class from bad file
        """
        bad_class = cls() # initialize class
        bad_class.read_bad_file(filename)
        return bad_class # return class as it is a constructor

class Bad(CoreBad):
    """
    Main class for bad
    """

    def __init__(self):
        """default constructor"""
        self.data = pd.DataFrame({"theta": np.empty([0])})

    @staticmethod
    def bad_BAB(atom, A, B, nl):
        """Return BAD distribution of B-A-B angle.

        Args:
            atom: ase atom object
            A, B: either int specifying atom numbers
                    A and B can be "X" (str) and in that case every specie will be taken into account
            nl: neighbor list used
        
        Returns:
            angles: list of floats representing B-A-B angles
                if B = "X", return X-A-Y angle. (all neighbours around A)
                if A and B = "X", return BAD distribution of X-Z-Y angle. (all neighbours around Z, for every Z)
        """    
        atomic_numbers = atom.get_atomic_numbers()  
        angles = [] #one set of 3 atoms will be comprised 1 time
        for a in range(len(atomic_numbers)):
            if A == "X" or atomic_numbers[a] == A:
                indices = nl[a]
                # indices, offsets = nl.get_neighbors(a)
                B_neighbors = [i for i in indices if B == "X" or atomic_numbers[i] == B] #Take couples of B-B neighbours
                comb = itertools.combinations(B_neighbors, 2) # Get all combinations of Bs
                
                angles_indices = []
                for pair in list(comb): 
                    i, j = pair
                    angles_indices.append([i,a,j])
    
                if angles_indices != []: # get_angles() doesn't work with an empty list
                    angles += list(atom.get_angles(angles_indices, mic=True))
        return angles
    
    def compute_bad_for_frame(self, atom, cutoff_dict, elements):
        """
        compute bad for ase atom object
        """
        # atom = trajectory[i]
        nl = amatom.get_neighborlist(atom, cutoff_dict)
        dic = {}
        for A, B in elements:
            aba_str = "-".join([ase.data.chemical_symbols[C] for C in [B, A, B]])
            dic[aba_str] = self.bad_BAB(atom, A, B, nl)  

        return dic
    
    def compute_bad(self, trajectory, nb_set_and_cutoff, dtheta, normalization, parallel):
        """
        compute compute_bad from ase trajectory object

        normalization not used so far, only one option (ie 'total')
        """
        atomic_numbers_unique = list(set(trajectory[0].get_atomic_numbers()))

        cutoff_dict = amatom.format_cutoff(nb_set_and_cutoff)
        elements_present_unique =  list(set([ase.data.atomic_numbers[i] for nb_set in nb_set_and_cutoff.keys() for i in nb_set.split('-') ]))

        if len(elements_present_unique) == len(atomic_numbers_unique):
            elements_present_unique.append("X")
        N_species = len(elements_present_unique) # number of different chemical species
            # N_species += 1 # include "X"

        # Partial RDFs               
        elements = [(a, b) for b in elements_present_unique for a in elements_present_unique if (a not in [b, "X"] or ((a, b) == ("X", "X"))) ]

        # nl = ase.neighborlist.NeighborList(cutoff_dict, self_interaction = False, bothways = True)
        # cutoffs = ase.neighborlist.natural_cutoffs(atom, mult = 1.2) # Generate a radial cutoff for every atom based on covalent radii ; mult is a multiplier for all cutoffs
        # nl = ase.neighborlist.NeighborList(cutoffs, self_interaction = False, bothways = True)

        rmax = np.max(list(nb_set_and_cutoff.values()))

        logger.info("Start computing bad for %s frames with dtheta = %s", len(trajectory), dtheta)
        bins = int(180 // dtheta)
        theta_bins = np.arange(bins + 2) * dtheta    
        theta = np.arange(bins + 1) * dtheta + dtheta / 2   
        self.data = pd.DataFrame({"theta": theta})    


        if parallel == False:
            list_of_dict = [self.compute_bad_for_frame(trajectory[i], cutoff_dict, elements) for i in range(len(trajectory))]
        else:
            num_cores = parallel if type(parallel) == int else max(joblib.cpu_count() // 2 - 2, 2) # heuristic for 40cores Xeon cpus
            list_of_dict = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(self.compute_bad_for_frame)(trajectory[i], cutoff_dict, elements) for i in range(len(trajectory)))

        aba_str_keys = list_of_dict[0].keys()
        for aba_str in aba_str_keys:
            angles = []
            for dic in list_of_dict:
                angles += dic[aba_str]
            if angles != []:
                self.data[aba_str] = np.histogram(angles, bins = theta_bins, density=True)[0]


    def write_to_file(self, filename):
        filename = amof.files.path.append_suffix(filename, 'bad')
        self.data.to_feather(filename)

    def read_bad_file(self, path_to_data):
        path_to_data = amof.files.path.append_suffix(path_to_data, 'bad')
        self.data = pd.read_feather(path_to_data)


class BadByCn(CoreBad):
    """
    Modified Bad class that computes BAD by Coordination Number (BAD for A bonded to 1 B, 2 B, etc.)

    Partial BAD by cn are normalized such that the sum over cn of partial BAD[cn] is of total area 1 (same as Bad class result).
    Each cn has a weight proportional to the number of angles found for A atoms with cn B nn.

    Limited derivation from CoreBad class as the data structure behind is xarray instead of pandas
    Bad could later be adapted with this new datastructures if this class end up routinely used
    """

    def __init__(self):
        """default constructor"""
        self.data = xr.DataArray(np.empty([0,0,0]), 
            coords = [('theta', np.empty([0], dtype='float64')), # one of numpy float
                ('atom_triple', np.empty([0], dtype='str_')), 
                ('cn', np.empty([0], dtype='str_'))], # numpy str is str_
            name = 'bad')

    @staticmethod
    def bad_BAB(atom, A, B, nl):
        """Return BAD distribution of B-A-B angle.

        Args:
            atom: ase atom object
            A, B: either int specifying atom numbers
                    A and B can be "X" (str) and in that case every specie will be taken into account
            nl: neighbor list used
        
        Returns:
            angles: dict of list of floats representing B-A-B angles
                Each dict key is the number of B nb of A
                if B = "X", return X-A-Y angle. (all neighbours around A)
                if A and B = "X", return BAD distribution of X-Z-Y angle. (all neighbours around Z, for every Z)
        """    
        atomic_numbers = atom.get_atomic_numbers()  
        angles = {} #one set of 3 atoms will be comprised 1 time
        for a in range(len(atomic_numbers)):
            if A == "X" or atomic_numbers[a] == A:
                indices = nl[a]
                # indices, offsets = nl.get_neighbors(a)
                B_neighbors = [i for i in indices if B == "X" or atomic_numbers[i] == B] #Take couples of B-B neighbours
                cn = len(B_neighbors)
                if cn > 1 and cn not in angles.keys():
                    angles[cn] = []

                comb = itertools.combinations(B_neighbors, 2) # Get all combinations of Bs
                angles_indices = []
                for pair in list(comb): 
                    i, j = pair
                    angles_indices.append([i,a,j])
    
                if angles_indices != []: # get_angles() doesn't work with an empty list
                    angles[cn] += list(atom.get_angles(angles_indices, mic=True))
        return angles
    
    def compute_bad_for_frame(self, atom, cutoff_dict, elements):
        """
        compute bad for ase atom object
        """
        # atom = trajectory[i]
        nl = amatom.get_neighborlist(atom, cutoff_dict)
        dic = {}
        for A, B in elements:
            aba_str = "-".join([ase.data.chemical_symbols[C] for C in [B, A, B]])
            dic[aba_str] = self.bad_BAB(atom, A, B, nl)  

        return dic
    
    def compute_bad(self, trajectory, nb_set_and_cutoff, dtheta, normalisation, parallel):
        """
        compute compute_bad from ase trajectory object
        """
        atomic_numbers_unique = list(set(trajectory[0].get_atomic_numbers()))

        cutoff_dict = amatom.format_cutoff(nb_set_and_cutoff)
        elements_present_unique =  list(set([ase.data.atomic_numbers[i] for nb_set in nb_set_and_cutoff.keys() for i in nb_set.split('-') ]))

        if len(elements_present_unique) == len(atomic_numbers_unique):
            elements_present_unique.append("X")
        N_species = len(elements_present_unique) # number of different chemical species
            # N_species += 1 # include "X"

        # Partial RDFs               
        elements = [(a, b) for b in elements_present_unique for a in elements_present_unique if (a not in [b, "X"] or ((a, b) == ("X", "X"))) ]

        # nl = ase.neighborlist.NeighborList(cutoff_dict, self_interaction = False, bothways = True)
        # cutoffs = ase.neighborlist.natural_cutoffs(atom, mult = 1.2) # Generate a radial cutoff for every atom based on covalent radii ; mult is a multiplier for all cutoffs
        # nl = ase.neighborlist.NeighborList(cutoffs, self_interaction = False, bothways = True)

        rmax = np.max(list(nb_set_and_cutoff.values()))

        logger.info("Start computing bad for %s frames with dtheta = %s", len(trajectory), dtheta)
        bins = int(180 // dtheta)
        theta_bins = np.arange(bins + 2) * dtheta    
        theta = np.arange(bins + 1) * dtheta + dtheta / 2   
        # self.data = pd.DataFrame({"theta": theta})    


        if parallel == False:
            list_of_dict = [self.compute_bad_for_frame(trajectory[i], cutoff_dict, elements) for i in range(len(trajectory))]
        else:
            num_cores = parallel if type(parallel) == int else max(joblib.cpu_count() // 2 - 2, 2) # heuristic for 40cores Xeon cpus
            list_of_dict = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(self.compute_bad_for_frame)(trajectory[i], cutoff_dict, elements) for i in range(len(trajectory)))

        dic_of_xarray = {}
        aba_str_keys = list_of_dict[0].keys()
        for aba_str in aba_str_keys:
            angles_all = {}
            for dic_aba in list_of_dict:
                for cn, angles in dic_aba[aba_str].items():
                    if cn not in angles_all.keys():
                        angles_all[cn] = []
                    angles_all[cn] += angles
            if angles_all != {}:
                if normalisation == 'partial': 
                    num_angles_all = np.sum(len(ang) for ang in angles_all.values())
                cn_coord, bad_list = [], []
                for cn, ang in angles_all.items():
                    cn_coord.append(cn)
                    normalization_ratio = len(ang)/num_angles_all if normalisation == 'partial' else 1
                    bad_list.append(normalization_ratio * np.histogram(angles_all[cn], bins = theta_bins, density=True)[0])
                ar = xr.DataArray(bad_list, coords={"cn":cn_coord, "theta": theta})
                dic_of_xarray[aba_str] = ar
                1
        xa = xr.Dataset(dic_of_xarray)
        xa = xa.to_array("atom_triple")
        xa = xr.Dataset({'bad': xa}) 
        self.data = xa


    def write_to_file(self, filename):
        filename = amof.files.path.append_suffix(filename, 'bad')
        self.data.to_netcdf(filename)

    def read_bad_file(self, filename):
        filename = amof.files.path.append_suffix(filename, 'bad')
        self.data = xr.open_dataset(filename)



"""
Module for representing symbols not present in the periodic table

Designed for fragment names, but more general terminology:
    name (e.g. fragname such as Im)
    symbol (chemical_symbol, from the periodic table)
"""

import amof.files.path
import json

from ase.data import chemical_symbols
# Seventh line of the periodic table (from ase.data)
# Arbitrary list for now with elements unlikely to be present in ZIFs, need to be expanded if not long enough
chemical_symbols_seventh_period = ['Fr', 'Ra', 'Ac', 'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk',
    'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr',
    'Rf', 'Db', 'Sg', 'Bh', 'Hs', 'Mt', 'Ds', 'Rg', 'Cn', 'Nh', 'Fl', 'Mc',
    'Lv', 'Ts', 'Og']    

class DummySymbols(object):
    """
    Class allowing handling of any name represented by a string by linking it to a symbol present in the periodic table
    """

    def __init__(self, names = None):
        """
        Args:
            symbol_names: list of str, names used in an atom/structure object
                if None, will create empty class
        """ 
        self.from_name_to_symbol = {}     
        self.from_symbol_to_name = {}
        self.names = [] # list of names
        self.nb_changed_names = 0  
        self.available_chemical_symbols = chemical_symbols_seventh_period
        if names is not None:
            self.add_names(names)

    def add_names(self, names):
        """
        Add names if not already present

        Args:
            symbol_names: list of str, names used in an atom/structure object
                if None, will create empty class
        """ 
        new_names = [n for n in names if n not in self.names]
        for name in new_names:
            if name in chemical_symbols:
                pt_symbol = name
                if name in self.available_chemical_symbols:
                    self.available_chemical_symbols.pop(name)
            else:
                pt_symbol = self.available_chemical_symbols[self.nb_changed_names] # will raise error if insufficient available names
                self.nb_changed_names += 1
            self.from_name_to_symbol[name] = pt_symbol
            self.from_symbol_to_name[pt_symbol] = name
            self.names.append(name)

    def get_symbol(self, name):
        return self.from_name_to_symbol[name]
    
    def get_name(self, symbol):
        return self.from_symbol_to_name[symbol]
    
    @classmethod
    def from_file(cls, filename):
        """
        Constructor of class from file
        """
        cn_class = cls() # initialize class
        cn_class.read_file(filename)
        return cn_class # return class as it is a constructor

    def read_file(self, filename):
        """filename: str"""
        filename = amof.files.path.append_suffix(filename, 'symbols')
        self.from_name_to_symbol = json.load(open(filename))
        self.from_symbol_to_name = {v: k for k, v in self.from_name_to_symbol.items()}
        self.names = list(self.from_name_to_symbol.keys())
        self.nb_changed_names = sum([v == k for k, v in self.from_name_to_symbol.items()])
        self.available_chemical_symbols = [s for s in self.available_chemical_symbols if s not in self.names]

    def write_to_file(self, filename):
        filename = amof.files.path.append_suffix(filename, 'symbols')
        with open(filename, 'w') as fp:
            json.dump(self.from_name_to_symbol, fp)

    def __str__(self):
        return ', '.join([':'.join([k, v]) for k, v in self.from_name_to_symbol.items()])



"""
amof.coordination
---
Submodule containing coordination analysis related methods
Import entire core content
Called by every submodule in coordination
"""

# from . import core # to import core when importing submodule
from .core import *


"""
Module containing pore analysis related methods
"""

from subprocess import TimeoutExpired
import ase
import ase.data

import numpy as np
import pandas as pd
import logging
import tempfile
import re
import joblib

import amof.trajectory
import amof.atom
import amof.files.path
import amof.pore.pysimmzeopp

logger = logging.getLogger(__name__)

class Pore(object):
    """
    Main class for pore analysis
    """

    def __init__(self):
        """default constructor"""
        self.data = pd.DataFrame({"Step": np.empty([0])})

    @classmethod
    def from_trajectory(cls, trajectory, delta_Step = 1, first_frame = 0, parallel = False):
        """
        constructor of pore class from an ase trajectory object

        """
        pore_class = cls() # initialize class
        step = amof.trajectory.construct_step(delta_Step=delta_Step, first_frame = first_frame, number_of_frames = len(trajectory))
        pore_class.compute_surface_volume(trajectory, step, parallel)
        return pore_class # return class as it is a constructor

    def compute_surface_volume(self, trajectory, step, parallel):
        """
        Args:
            trajectory: ase trajectory object
            step: np array, simulation steps
            parallel: Boolean or int (number of cores to use): whether to parallelize the computation
        """
        logger.info("Start pore analysis for volume and surfaces for %s frames", len(trajectory))
        list_of_dict = []
        if parallel == False:
            for i in range(len(trajectory)):
                logger.debug('compute frame # %s out of %s', i + 1, len(trajectory))
                list_of_dict.append(self.get_surface_volume(trajectory[i], step[i]))
        else:
            if type(parallel) == int:
                num_cores = parallel
            else:
                num_cores = max(joblib.cpu_count() // 2 - 2, 2) # heuristic for 40cores Xeon cpus: less than 20 and nice value for 50 steps
            list_of_dict = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(self.get_surface_volume)(trajectory[i], step[i]) for i in range(len(trajectory)))

        list_of_dict = [dic for dic in list_of_dict if dic is not None] # filter pore volume that are not properly computed

        if list_of_dict != []:
            df = pd.DataFrame(list_of_dict)
            self.data = df 
        # else keep empty surface_volume from init

    @staticmethod
    def read_zeopp(filename):
        """
        read surface (sa) and volume (vol) zeopp output
        """
        with open(filename) as f:
            first_line = f.readline().strip('\n')
        split_line = re.split('\ +', first_line)
        split_line = split_line[6:] # remove file name, density and unit cell volume
        list_of_keys = [s.strip(':') for s in split_line[::2]]
        list_of_values = [float(s) for s in split_line[1::2]]
        dic = dict(zip(list_of_keys, list_of_values))
        return dic

    def get_surface_volume(self, atom, step):
        """
        Args:
            atom: ase atom object
            step: int, represent step of frame
        Returns:
            dic: dictionary with output from zeopp vol and sa
        """
        with tempfile.TemporaryDirectory() as tmpdirname:
            atom.write(tmpdirname + '/atom.cif')
            try:
                amof.pore.pysimmzeopp.network(tmpdirname + '/atom.cif', sa = True, vol = True)
                sa = self.read_zeopp(tmpdirname + '/atom.sa')
                vol = self.read_zeopp(tmpdirname + '/atom.vol')
                dic = {'Step': step, **sa, **vol}
            except TimeoutExpired:
                logger.warning('Timout expired for ZEOpp. System size: %s; Step: %s', atom.get_global_number_of_atoms(), step)
                dic = None
        return dic

    def write_to_file(self, filename):
        """path_to_output: where the pore object will be written"""
        filename = amof.files.path.append_suffix(filename, 'pore')
        self.data.to_feather(filename)

    @classmethod
    def from_file(cls, filename):
        """
        constructor of pore class from msd file
        """
        pore_class = cls() # initialize class
        pore_class.read_surface_volume_file(filename)
        return pore_class # return class as it is a constructor

    def read_surface_volume_file(self, filename):
        """path_to_data: where the MSD object is"""
        filename = amof.files.path.append_suffix(filename, 'pore')
        self.data = pd.read_feather(filename)



"""
Module containing ring statistics analysis related methods

Terminology: 
    singular form 'ring' is employed by default to talk about ring statistics analysis
    plural form 'rings' or 'RINGS' refers to the Rigorous Investigation of Networks Generated using Simulations or ‘RINGS’ code
"""

import ase
import ase.data

import numpy as np
import pandas as pd
import logging
import os
import pathlib
import tempfile
import re
import joblib
import atomman
import itertools
import subprocess
import shlex
import xarray as xr

import amof.trajectory
import amof.atom as amatom
import amof.files.path as ampath
import amof.pore.pysimmzeopp

logger = logging.getLogger(__name__)

class Ring(object):
    """
    Main class for ring statistics analysis analysis

    Search for primitive rings as defined by
        Le Roux, S., & Jund, P. (2010). Ring statistics analysis of topological networks: 
        New approach and application to amorphous GeS2 and SiO2 systems. 
        Computational Materials Science, 49(1), 70–83. 
        https://doi.org/10.1016/j.commatsci.2010.04.023 

    A definition phrased by Franzblau of such rings (named SP ring in this paper) is:
        Given a graph G and a ring R in 6, R is a shortest path ring (SP ring) 
        if R contains a shortest path for each pair of vertices in the ring. 
        That is, distG(y, z ) =dist+ (y, z) for each pair y, z in the ring

        Franzblau, D. S. (1991). Computation of ring statistics for network models of solids. 
        Physical Review B, 44(10), 4925–4930. https://doi.org/10.1103/PhysRevB.44.4925
    """

    def __init__(self, max_search_depth = None, discard_if_potentially_undiscovered_rings = False):
        """default constructor"""
        self.data = xr.DataArray(np.empty([0,0,0]), 
            coords = [('Step', np.empty([0], dtype='int64')), 
                ('ring_size', np.empty([0], dtype='int64')), 
                ('ring_var', np.empty([0], dtype='str_'))], # numpy str is str_
            name = 'ring')
        self.max_search_depth = max_search_depth
        self.discard_if_potentially_undiscovered_rings = discard_if_potentially_undiscovered_rings
        self.report_search = pd.DataFrame({"Step": np.empty([0])})

    @classmethod
    def from_trajectory(cls, trajectory, nb_set_and_cutoff, max_search_depth = 32 , delta_Step = 1, first_frame = 0, parallel = False):
        """
        Constructor of ring class from an ase trajectory object

        Args:
            nb_set_and_cutoff: dict, keys are str indicating pair of neighbours, 
                values are cutoffs float, in Angstrom
            max_search_depth: int, maximum search depth for rings.
        """
        ring_class = cls(max_search_depth = max_search_depth) # initialize class
        nb_set_and_cutoff_list = [nb_set_and_cutoff for i in range(len(trajectory))]
        step = amof.trajectory.construct_step(delta_Step=delta_Step, first_frame = first_frame, number_of_frames = len(trajectory))
        ring_class.compute_ring(trajectory, nb_set_and_cutoff_list, step, parallel)
        return ring_class # return class as it is a constructor

    @classmethod
    def from_reduced_trajectory(cls, reduced_trajectory, max_search_depth = 32, 
            discard_if_potentially_undiscovered_rings = False, parallel = False):
        """
        Constructor of ring class from a amof ReducedTrajectory object
        
        Args:
            reduced_trajectory: amof ReducedTrajectory object
            nb_set_and_cutoff: dict, keys are str indicating pair of neighbours, 
                values are cutoffs float, in Angstrom
        """
        ring_class = cls(max_search_depth = max_search_depth,
            discard_if_potentially_undiscovered_rings = discard_if_potentially_undiscovered_rings) # initialize class with empty data
        criteria_to_compute_ring = ['connectivity_constructible_with_cutoffs'] # can be expanded in further dev
        criteria_enlarged = ['in_reduced_trajectory'] + criteria_to_compute_ring
        rs = reduced_trajectory.report_search
        rs_traj = rs[rs['in_reduced_trajectory']==True]
        if (len(rs_traj) != 0 and
            np.min([criterium in rs_traj.columns for criterium in criteria_to_compute_ring]) == True):
            compute_ring = rs[criteria_enlarged].all(axis='columns')
            if np.sum(compute_ring) != 0:
                subset_reduced_traj = rs_traj[criteria_to_compute_ring].all(axis='columns')
                nb_set_and_cutoff_list = [eval(i) for i in rs[compute_ring]['nb_set_and_cutoff']]
                step = np.array(rs[compute_ring].index)
                traj = list(itertools.compress(reduced_trajectory.trajectory, subset_reduced_traj))
                ring_class.compute_ring(traj, nb_set_and_cutoff_list, step, parallel)
                return ring_class

        logger.info('No valid frame in reduced trajectory')
        return ring_class # return empty class 

    def compute_ring(self, trajectory, nb_set_and_cutoff_list, step, parallel):
        """
        Args:
            trajectory: ase trajectory object
                nb_set_and_cutoff_list: list of dict, one per frame in trajectory
            step: np array, simulation steps
            parallel: Boolean or int (number of cores to use): whether to parallelize the computation
        """
        logger.info("Start ring analysis for %s frames", len(trajectory))
        result_list = []

        if parallel == False:
            for i in range(len(trajectory)):
                logger.debug('compute frame # %s out of %s', i + 1, len(trajectory))
                result_list.append(self.compute_ring_for_atom(trajectory[i], step[i], nb_set_and_cutoff_list[i]))
        else:
            if type(parallel) == int:
                num_cores = parallel
            else:
                num_cores = max(joblib.cpu_count() // 2 - 2, 2) # heuristic for 40cores Xeon cpus: less than 20 and nice value for 50 steps
            result_list = joblib.Parallel(n_jobs=num_cores)(joblib.delayed(self.compute_ring_for_atom)(
                        trajectory[i], step[i], nb_set_and_cutoff_list[i]) for i in range(len(trajectory)))

        list_report_search = []
        list_of_xarray = []
        for ring_ar, report_search in result_list:
            list_report_search.append(report_search)
            if ring_ar is not None:
                list_of_xarray.append(ring_ar)

        self.report_search = pd.DataFrame(list_report_search).set_index('Step')

        if list_of_xarray != []:
            dic_of_xarray = dict(zip(step, list_of_xarray))
            xa = xr.Dataset(dic_of_xarray)
            xa = xa.to_array("Step", "Step")
            xa = xr.Dataset({'ring': xa}) # one large data_array containing thermo variables
            # xa = xa.to_dataset(dim='thermo') # separate thermo variables
            xa = xa.fillna(0) #  for every ring size found in the entire traj, add 0 at every step if none detected while other ring sizes are
            self.data = xa

    def read_rings_output(self, rstat_path):
        """
        read RINGS output from RINGS-res-5.dat, corresponding to primitive rings 
        read RINGS-res-3.dat, corresponding to King's shortest path rings, to know whether some rings are missing

        Args:
            rstat_path: pathlib path leading to rstat, containing evol-RINGS files

        Returns:
            ar: xarray dataarray containing the rings info
            potentially_undiscovered_rings: float, number potentially_undiscovered_rings 
                    specified in RINGS output files. If 0, means that all are discovered.
        """
        filename = 'RINGS-res-3.dat' # King's shortest path rings
        with open(rstat_path / filename) as f:
            first_line = f.readline().strip('\n')
        searchObj = re.search(r'# Number of rings with n >  (.*) nodes which potentialy exist: (.*)', first_line, re.M|re.I)
        potentially_undiscovered_rings = round(float(searchObj.group(2)))
        
        filename = 'RINGS-res-5.dat' # primitive rings 
        header = 1
        df = pd.read_csv(rstat_path / filename, header = header, escapechar='#', sep='\s+')
        df = df.set_index(' n')
        ar = xr.DataArray(df, dims=("ring_size", "ring_var"))
        return ar, potentially_undiscovered_rings

    @staticmethod
    def fill_template(template_name, parameters, path):
        """
        Read, fill and write template

        Args:
            template_name: name of template file used in .ini file
            parameters: dic containing all the necessary input for templating
            path: pathlib path, where to write filled template under the name template_name
        """
        with open(pathlib.Path(__file__).parent.resolve() / 'template' / template_name) as f:
            template = f.read()
        script = atomman.tools.filltemplate(template, parameters, '{', '}')  
        with open(path / template_name, 'w') as f:
            f.write(script)  

    def write_input_files(self, atom, cutoff_dict, search_depth, path):
        """
        Write RINGS input and option files in path

        Args:
            atom: ase atom object
            cutoff_dict: 
            search_depth: int, rings_maximum_search_depth
            path: pathlib path
        """
        parameters = {}
        parameters['number_of_atoms'] = atom.get_global_number_of_atoms()
        atomic_numbers_unique = list(set(atom.get_atomic_numbers()))
        elements_present_unique =  [ase.data.chemical_symbols[i] for i in atomic_numbers_unique]
        parameters['number_of_chemical_species'] = len(elements_present_unique)
        parameters['list_of_chemical_species'] = ' '.join(elements_present_unique)
        parameters['rings_maximum_search_depth_divided_by_two'] = search_depth // 2 
        for i in range(3):
            parameters[f"cell{'abc'[i]}"] = str(atom.cell[i])[1:-1]
        parameters['cutoff_lines'] = ''
        for key, value in cutoff_dict.items():
            line_list = [ase.data.chemical_symbols[i] for i in key] + [str(value), """# \n """]
            parameters['cutoff_lines'] += ' '.join(line_list)
        parameters['Grtot'] = max(cutoff_dict.values())
        self.fill_template('input.inp', parameters, path)
        self.fill_template('options', parameters, path)

    def compute_ring_for_atom(self, atom, step, nb_set_and_cutoff):
        """
        Args:
            atom: ase atom object
            
        Returns:
            ring_ar: xarray dataarray containing the rings info
            report_search: dic
        """
        report_search = {'Step': step, 
            'Discarded frame': False, 
            'max_search_depth': self.max_search_depth,
            'Discard if potentially undiscovered rings': self.discard_if_potentially_undiscovered_rings,
            'Rings statistics computed with potentially undiscovered rings': False}
        cutoff_dict = amatom.format_cutoff(nb_set_and_cutoff, sort_pair = True)
        # add 0 where cutoff is not defined
        atomic_numbers_unique = list(set(atom.get_atomic_numbers()))
        for pair in itertools.combinations_with_replacement(atomic_numbers_unique, 2):
            if pair not in cutoff_dict.keys():
                cutoff_dict[pair] = 0
        # enlarge cutoff dict with 0 for every other values

        with tempfile.TemporaryDirectory() as tmpdirname:
            tmpdirpath = pathlib.Path(tmpdirname)
            (tmpdirpath / 'data').mkdir(parents=True, exist_ok=True)
            atom.write(tmpdirname + '/data/atom.xyz')
            
            # warning, make sure no user input is possible inside arg. 
            # Here shlex.quote is used as precaution in the only possible variable of the string
            arg = f"cd {shlex.quote(tmpdirname)} && rings input.inp"

            search_depth = min(16, self.max_search_depth)

            ring_ar = None
            potentially_undiscovered_rings = np.inf
            while search_depth <= self.max_search_depth and potentially_undiscovered_rings > 0:
                self.write_input_files(atom, cutoff_dict, search_depth, tmpdirpath)
                
                p = subprocess.Popen(arg, shell=True, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE) 
                p.wait()

                ring_ar, potentially_undiscovered_rings = self.read_rings_output(tmpdirpath / 'rstat')
                report_search['Final search_depth'] = search_depth
                report_search['Potentially undiscovered rings'] = potentially_undiscovered_rings

                search_depth += 4
            if potentially_undiscovered_rings > 0:
                logger.warning('Rings with n >  %s nodes potentialy exist', self.max_search_depth)
                report_search['Rings statistics computed with potentially undiscovered rings'] = True
                if self.discard_if_potentially_undiscovered_rings == True:
                    report_search['Discarded frame'] = True
                    ring_ar = None # don't add this frame to rings file
        return ring_ar, report_search

    def write_to_file(self, filename):
        """path_to_output: where the ring object will be written"""
        self.data.to_netcdf(ampath.append_suffix(filename, 'ring'))
        self.report_search.to_csv(ampath.append_suffix(filename, 'report_search.csv'))


    @classmethod
    def from_file(cls, filename):
        """
        constructor of ring class from ring file
        """
        ring_class = cls() # initialize class
        ring_class.read_ring_file(filename)
        return ring_class # return class as it is a constructor

    def read_ring_file(self, filename):
        """path_to_data: where the ring object is"""
        filename = ampath.append_suffix(filename, 'ring')
        self.data = xr.open_dataset(filename)



"""
Module containing basic file operation methods
"""

import gzip
import shutil
import os
import pathlib
import logging

logger = logging.getLogger(__name__)

def compress(filename, remove_if_exists = False):
    """
    Args:
        filename: str, filename without '.gz'
        remove_if_exists: bool, remove uncompressed file if True and a file with .gz already exist
    """
    if not (remove_if_exists == True and pathlib.Path(filename+'.gz').exists()):
        logger.info("compress %s", filename)
        with open(filename, 'rb') as f_in:
            with gzip.open(filename + '.gz', 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)
    os.remove(filename)

def decompress(filename, remove = True):
    """
    Args:
        filename: str, filename without '.gz'
        remove: bool, remove compressed file if True
    """
    logger.info("decompress %s", filename)
    with gzip.open(filename + '.gz', 'rb') as f_in:
        with open(filename, 'wb') as f_out:
            shutil.copyfileobj(f_in, f_out)
    if remove == True:
        os.remove(filename+ '.gz')

def concatenate(filenames, output_file):
    """
    Args:
        filename: list of strings
        output files: strings
    """
    with open(output_file,'wb') as wfd:
        for f in filenames:
            with open(f,'rb') as fd:
                shutil.copyfileobj(fd, wfd)